#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"M�dulo de Intefase para archivos de texto (bono fiscal version 1)"

__author__ = "Mariano Reingart (mariano@nsis.com.ar)"
__copyright__ = "Copyright (C) 2009 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.18a"

import datetime
import os
import sys
import time
import traceback
from ConfigParser import SafeConfigParser

# revisar la instalaci�n de pyafip.ws:
import wsaa
import wsbfev1


HOMO = False
DEBUG = False
XML = False
CONFIG_FILE = "rece.ini"

LICENCIA = """
receb.py: Interfaz de texto para generar Facturas Electr�nicas Bienes de
Capital
Copyright (C) 2008/2009 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para informaci�n adicional sobre garant�a, soporte t�cnico comercial
e incorporaci�n/distribuci�n en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

# definici�n del formato del archivo de intercambio:
N = 'Numerico'
A = 'Alfanumerico'
I = 'Importe'
ENCABEZADO = [
    ('tipo_reg', 1, N),  # 0: encabezado
    ('fecha_cbte', 8, A),
    ('tipo_cbte', 2, N), ('punto_vta', 4, N),
    ('cbte_nro', 8, N),
    ('tipo_doc', 2, N), ('nro_doc', 11, N),
    ('imp_total', 15, I), ('imp_tot_conc', 15, I),
    ('imp_neto', 15, I), ('impto_liq', 15, I),
    ('impto_liq_rni', 15, I), ('imp_op_ex', 15, I),
    ('impto_perc', 15, I), ('imp_iibb', 15, I),
    ('impto_perc_mun', 15, I), ('imp_internos', 15, I),
    ('imp_moneda_id', 3, A),
    ('imp_moneda_ctz', 10, I),
    ('zona', 5, A),
    ('cae', 14, N), ('fecha_vto', 8, A),
    ('resultado', 1, A), ('obs', 2, A), ('reproceso', 1, A),
    ('id', 15, N),
    ]

DETALLE = [
    ('tipo_reg', 1, N),  # 1: detalle item
    ('ncm', 15, A),
    ('sec', 15, A),
    ('qty', 15, I),
    ('umed', 5, N),
    ('precio', 15, I),
    ('bonif', 15, I),
    ('imp_total', 15, I),
    ('iva_id', 5, N),
    ('ds', 200, A),
    ]


def leer(linea, formato):
    dic = {}
    comienzo = 1
    for (clave, longitud, tipo) in formato:
        valor = linea[comienzo-1:comienzo-1+longitud].strip()
        if tipo == N and valor:
            valor = str(int(valor))
        if tipo == I:
            if valor:
                valor = float("%s.%02d" % (int(valor[:-2]), int(valor[-2:])))
            else:
                valor = 0.00
        dic[clave] = valor

        comienzo += longitud
    return dic


translate_keys = {
    'ncm': 'Pro_codigo_ncm', 'bonif': 'Imp_bonif', 'precio': 'Pro_precio_uni',
    'sec': 'Pro_codigo_sec', 'ds': 'Pro_ds', 'umed': 'Pro_umed',
    'qty': 'Pro_qty', 'imp_moneda_id': 'Imp_moneda_Id'
}


def escribir(dic, formato):
    linea = " " * 335
    comienzo = 1
    for (clave, longitud, tipo) in formato:
        if clave.capitalize() in dic:
            clave = clave.capitalize()
        valor = str(dic.get(clave, ""))
        if valor == "" and clave in translate_keys:
            valor = str(dic.get(translate_keys[clave], ""))
        if tipo == N and valor and valor != "NULL":
            valor = ("%%0%dd" % longitud) % int(valor)
        elif tipo == I and valor:
            valor = ("%%0%dd" % longitud) % (float(valor)*100)
        else:
            valor = ("%%0%ds" % longitud) % valor
        linea = linea[:comienzo-1] + valor + linea[comienzo-1+longitud:]
        comienzo += longitud
    return linea + "\n"


def autorizar(ws, entrada, salida):
    # recupero el �ltimo n�mero de transacci�n
    # id = wsbfe.ultnro(client, token, sign, cuit)

    detalles = []
    encabezado = {}
    for linea in entrada:
        if str(linea[0]) == '0':
            encabezado = leer(linea, ENCABEZADO)
        elif str(linea[0]) == '1':
            detalle = leer(linea, DETALLE)
            detalles.append(detalle)
        else:
            print("Tipo de registro incorrecto:", linea[0])

    if not encabezado['id'].strip():
        # TODO: habria que leer y/o grabar el id en el archivo
        # id += 1 # incremento el n� de transacci�n
        # Por el momento, el id se calcula con el tipo, pv y n� de comprobant
        i = long(encabezado['cbte_nro'])
        i += (
            int(encabezado['cbte_nro'])*10**4 + int(
                encabezado['punto_vta']))*10**8
        encabezado['id'] = i

    if not encabezado['zona'].strip():
        encabezado['zona'] = 0

    if 'testing' in sys.argv:
        ult_cbte = ws.GetLastCMP(punto_vta, tipo_cbte)
        encabezado['cbte_nro'] = ult_cbte + 1
        ult_id = ws.GetLastID()
        encabezado['id'] = ult_id + 1

    # encabezado['imp_moneda_ctz'] = 1.00
    ws.CrearFactura(**encabezado)
    for detalle in detalles:
        ws.AgregarItem(**detalle)

    if DEBUG:
        print('\n'.join(["%s='%s'" % (k, v) for k, v in ws.factura.items()]))
        print('id:', encabezado['id'])
    if not DEBUG or raw_input("Facturar?") == "S":
        cae = ws.Authorize(encabezado['id'])
        dic = ws.factura
        dic.update(
            {'id':  encabezado['id'],
             'fecha_cbte': ws.FechaCbte,
             'imp_total': ws.ImpTotal or 0,
             'imp_neto': ws.ImpNeto or 0,
             'impto_liq': ws.ImptoLiq or 0,
             'cae': str(ws.CAE),
             'obs': str(ws.Obs), 'reproceso': str(ws.Reproceso),
             'fch_venc_cae': ws.Vencimiento,
             'err_msg': ws.ErrMsg})
        escribir_factura(dic, salida)
        print(
            "ID:", dic['id'], "CAE:", dic['cae'], "Obs:", dic['obs'],
            "Reproceso:", dic['reproceso'])


def escribir_factura(dic, archivo):
    dic['tipo_reg'] = 0
    archivo.write(escribir(dic, ENCABEZADO))
    for it in dic['detalles']:
        it['tipo_reg'] = 1
        archivo.write(escribir(it, DETALLE))


def depurar_xml(client):
    fecha = time.strftime("%Y%m%d%H%M%S")
    f = open("request-%s.xml" % fecha, "w")
    f.write(client.xml_request)
    f.close()
    f = open("response-%s.xml" % fecha, "w")
    f.write(client.xml_response)
    f.close()


if __name__ == "__main__":
    if '/ayuda' in sys.argv:
        print(LICENCIA)
        print("Opciones: ")
        print(" /ayuda: este mensaje")
        print(" /dummy: consulta estado de servidores")
        print(
            " /prueba: genera y autoriza una factura de prueba \
            (no usar en producci�n!)")
        print(" /ult: consulta �ltimo n�mero de comprobante")
        print(" /id: consulta �ltimo ID")
        print(" /debug: modo depuraci�n (detalla y confirma las operaciones)")
        print(
            " /formato: muestra el formato de los archivos de entrada/salida")
        print(
            " /get: recupera datos de un comprobante autorizado previamente \
            (verificaci�n)")
        print(
            " /xml: almacena los requerimientos y respuestas XML \
            (depuraci�n)")
        print(
            "Ver rece.ini para par�metros de configuraci�n \
            (URL, certificados, etc.)")
        sys.exit(0)

    if '/debug'in sys.argv:
        DEBUG = True
        print("VERSION", __version__, "HOMO", HOMO)

    # si se pasa el archivo de configuraci�n por par�metro, confirmar que
    # exista
    # y descartar que sea una opci�n
    if len(sys.argv) > 1 and (
        sys.argv[1][0] not in "-/" or os.path.exists(sys.argv[1])
    ):
        CONFIG_FILE = sys.argv.pop(1)
    if DEBUG:
        print("CONFIG_FILE:", CONFIG_FILE)

    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    cert = config.get('WSAA', 'CERT')
    privatekey = config.get('WSAA', 'PRIVATEKEY')
    cuit = config.get('WSBFE', 'CUIT')
    entrada = config.get('WSBFE', 'ENTRADA')
    salida = config.get('WSBFE', 'SALIDA')

    if config.has_option('WSAA', 'URL') and not HOMO:
        wsaa_url = config.get('WSAA', 'URL')
    else:
        wsaa_url = wsaa.WSAAURL
    if config.has_option('WSBFE', 'URL') and not HOMO:
        wsbfe_url = config.get('WSBFE', 'URL')
    else:
        wsbfe_url = None

    if '/debug'in sys.argv:
        DEBUG = True

    if '/xml'in sys.argv:
        XML = True

    if DEBUG:
        print("wsaa_url %s\nwsbfe_url %s" % (wsaa_url, wsbfe_url))

    try:
        ws = wsbfev1.WSBFEv1()
        ws.Conectar("", wsbfe_url)
        ws.Cuit = cuit

        if '/dummy' in sys.argv:
            print("Consultando estado de servidores...")
            print(ws.Dummy())
            sys.exit(0)

        if '/formato' in sys.argv:
            print("Formato:")
            for msg, formato in [('Encabezado', ENCABEZADO),
                                 ('Detalle', DETALLE)]:
                comienzo = 1
                print("== %s ==" % msg)
                for (clave, longitud, tipo) in formato:
                    print(
                        " * Campo: %-20s Posici�n: %3d Longitud: %4d Tipo: \
                        %s" % (clave, comienzo, longitud, tipo)
                    )
                    comienzo += longitud
            sys.exit(0)

        # obteniendo el TA
        from wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wsbfe", cert, privatekey, wsaa_url)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)
        ws.SetTicketAcceso(ta)

        if '/prueba' in sys.argv or False:
            # generar el archivo de prueba para la pr�xima factura
            fecha = datetime.datetime.now().strftime("%Y%m%d")
            tipo_cbte = 1
            punto_vta = 2
            ult_cbte = int(ws.GetLastCMP(tipo_cbte, punto_vta)) + 1
            ult_id = ws.GetLastID() + 1

            f_entrada = open(entrada, "w")

            f = ws.CrearFactura(
                punto_vta=punto_vta, cbte_nro=ult_cbte,
                imp_moneda_id='PES', imp_moneda_ctz=1,
                fecha_cbte=fecha,
                imp_neto="390.00", impto_liq="81.90"
                )
            ws.AgregarItem(
                umed=7, ncm='7308.10.00', sec='', ds='prueba', qty=2.0,
                precio=100.0, bonif=0.0, iva_id=5, imp_total="242.00"
            )
            ws.AgregarItem(
                umed=7, ncm='7308.20.00', sec='', ds='prueba 2', qty=4.0,
                precio=50.0, bonif=10.0, iva_id=5, imp_total="229.90"
            )

            dic = ws.factura
            dic['id'] = ult_id
            escribir_factura(dic, f_entrada)
            f_entrada.close()

        if '/ult' in sys.argv:
            print("Consultar ultimo numero:")
            tipo_cbte = int(raw_input("Tipo de comprobante: "))
            punto_vta = int(raw_input("Punto de venta: "))
            ult_cbte = ws.GetLastCMP(tipo_cbte, punto_vta)
            print("Ultimo numero: ", ult_cbte)
            print("Fecha: ", ws.FechaCbte)
            depurar_xml(ws.client)
            sys.exit(0)

        if '/id' in sys.argv:
            ult_id = ws.GetLastID()
            print("ID: ", ult_id)
            depurar_xml(ws.client)
            sys.exit(0)

        if '/get' in sys.argv:
            print("Recuperar comprobante:")
            tipo_cbte = int(raw_input("Tipo de comprobante: "))
            punto_vta = int(raw_input("Punto de venta: "))
            cbte_nro = int(raw_input("Numero de comprobante: "))
            cae = ws.GetCMP(tipo_cbte, punto_vta, cbte_nro)
            cbt = {
                'fecha_cbte': ws.FechaCbte,
                'imp_total': ws.ImpTotal or 0,
                'imp_neto': ws.ImpNeto or 0,
                'impto_liq': ws.ImptoLiq or 0,
                'cae': str(ws.CAE),
                'obs': str(ws.Obs), 'reproceso': str(ws.Reproceso),
                'fch_venc_cae': ws.Vencimiento,
                'err_msg': ws.ErrMsg,
            }
            for k, v in cbt.items():
                print("%s = %s" % (k, v))
            depurar_xml(ws.client)
            sys.exit(0)

        f_entrada = f_salida = None
        try:
            f_entrada = open(entrada, "r")
            f_salida = open(salida, "w")
            try:
                autorizar(ws, f_entrada, f_salida)
            except Exception as e:
                XML = True
                raise
        finally:
            if f_entrada is not None:
                f_entrada.close()
            if f_salida is not None:
                f_salida.close()
            if XML:
                depurar_xml(ws.client)
        sys.exit(0)

    except Exception as e:
        print(unicode(e).encode("ascii", "ignore"))
        if DEBUG:
            raise
        sys.exit(5)



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo para obtener c�digo de autorizaci�n de impresi�n o
electr�nico del web service WSFEX de AFIP (Factura Electr�nica Exportaci�n)
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.23f"

import sys
from php import date, SimpleXMLElement, SoapClient

WSFEXURL = "https://wswhomo.afip.gov.ar/wsfex/service.asmx"
# WSFEXURL = "https://servicios1.afip.gov.ar/wsfex/service.asmx"
SOAP_ACTION = 'http://ar.gov.afip.dif.fex/'
SOAP_NS = "http://ar.gov.afip.dif.fex/"


class FEXError(RuntimeError):
    "Clase para informar errores del WSFEX"
    def __init__(self, fexerror):
        self.code = str(fexerror.ErrCode)
        self.msg = unicode(fexerror.ErrMsg)

    def __unicode__(self):
        return u"FEXError %s: %s" % (self.code, self.msg)


def dummy(client):
    "Metodo dummy para verificacion de funcionamiento"
    response = client.FEXDummy()
    result = response.FEXDummyResult
    appserver = dbserver = authserver = None
    try:
        appserver = str(result.AppServer)
        dbserver = str(result.DbServer)
        authserver = str(result.AuthServer)
    except (RuntimeError, IndexError, AttributeError) as e:
        print(e)
        pass
    return {'appserver': appserver,
            'dbserver': dbserver,
            'authserver': authserver}


def get_param_mon(client, token, sign, cuit):
    "Recuperador de valores referenciales de codigos de Moneda"
    response = client.FEXGetPARAM_MON(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_MONResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_MONResult.FEXErr)

    mons = []  # monedas
    for m in response.FEXGetPARAM_MONResult.FEXResultGet.ClsFEXResponse_Mon:
        mon = {'id': str(m.Mon_Id), 'ds': str(m.Mon_Ds).decode('utf8'),
               'vig_desde': str(m.Mon_vig_desde),
               'vig_hasta': str(m.Mon_vig_hasta)}
        mons.append(mon)
    return mons


def get_param_tipo_cbte(client, token, sign, cuit):
    "Recuperador de valores referenciales de c�digos de Tipos de comprobante"
    response = client.FEXGetPARAM_Tipo_Cbte(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_Tipo_CbteResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_Tipo_CbteResult.FEXErr)

    cbtes = []  # tipos de comprobantes
    for c in response.FEXGetPARAM_Tipo_CbteResult.\
            FEXResultGet.ClsFEXResponse_Tipo_Cbte:
        cbte = {'id': int(c.Cbte_Id), 'ds': str(c.Cbte_Ds).decode('utf8'),
                'vig_desde': str(c.Cbte_vig_desde),
                'vig_hasta': str(c.Cbte_vig_hasta)}
        cbtes.append(cbte)
    return cbtes


def get_param_tipo_expo(client, token, sign, cuit):
    "Recuperador de valores referenciales de c�digos de Tipo de exportaci�n"
    response = client.FEXGetPARAM_Tipo_Expo(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_Tipo_ExpoResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_Tipo_ExpoResult.FEXErr)

    tipos = []  # tipos de exportaci�n
    for t in response.FEXGetPARAM_Tipo_ExpoResult.\
            FEXResultGet.ClsFEXResponse_Tex:
        tipo = {'id': int(t.Tex_Id), 'ds': str(t.Tex_Ds).decode('utf8'),
                'vig_desde': str(t.Tex_vig_desde),
                'vig_hasta': str(t.Tex_vig_hasta)}
        tipos.append(tipo)
    return tipos


def get_param_idiomas(client, token, sign, cuit):
    "Recuperador de valores referenciales de c�digos de Idiomas"
    response = client.FEXGetPARAM_Idiomas(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_IdiomasResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_IdiomasResult.FEXErr)

    tipos = []  # tipos de exportaci�n
    for t in response.FEXGetPARAM_IdiomasResult.\
            FEXResultGet.ClsFEXResponse_Idi:
        tipo = {'id': int(t.Idi_Id), 'ds': str(t.Idi_Ds).decode('utf8'),
                'vig_desde': str(t.Idi_vig_desde),
                'vig_hasta': str(t.Idi_vig_hasta)}
        tipos.append(tipo)
    return tipos


def get_param_umed(client, token, sign, cuit):
    "Recuperador de valores referenciales de Unidades de Medidas"
    response = client.FEXGetPARAM_UMed(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_UMedResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_UMedResult.FEXErr)

    umeds = []  # unidades de medida
    for u in response.FEXGetPARAM_UMedResult.FEXResultGet.ClsFEXResponse_UMed:
        try:
            umed = {'id': int(u.Umed_Id), 'ds': str(u.Umed_Ds).decode('utf8'),
                    'vig_desde': str(u.Umed_vig_desde),
                    'vig_hasta': str(u.Umed_vig_hasta)}
        except Exception as e:
            print(e)
            # <ClsFEXResponse_UMed xsi:nil="true"/> WTF!
            # import pdb; pdb.set_trace()
            # print(u)
            pass

        umeds.append(umed)
    return umeds


def get_param_dst_pais(client, token, sign, cuit):
    "Recuperador de valores referenciales de c�digos de Pa�ses"
    response = client.FEXGetPARAM_DST_pais(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_DST_paisResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_DST_paisResult.FEXErr)

    paises = []
    for p in response.FEXGetPARAM_DST_paisResult.\
            FEXResultGet.ClsFEXResponse_DST_pais:
        pais = {
            'codigo': int(p.DST_Codigo), 'ds': str(p.DST_Ds).decode('utf8')}
        paises.append(pais)
    return paises


def get_param_dst_cuit(client, token, sign, cuit):
    "Recuperador de valores referenciales de CUITs de Paises"
    response = client.FEXGetPARAM_DST_CUIT(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_DST_CUITResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_DST_CUITResult.FEXErr)

    paises = []
    for p in response.FEXGetPARAM_DST_CUITResult.\
            FEXResultGet.ClsFEXResponse_DST_cuit:
        pais = {'cuit': int(p.DST_CUIT), 'ds': str(p.DST_Ds).decode('utf8')}
        paises.append(pais)
    return paises


def get_param_ctz(client, token, sign, cuit, moneda_id):
    "Recuperador de cotizaci�n de moneda"
    response = client.FEXGetPARAM_Ctz(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)},
        Mon_id=moneda_id)

    if int(response.FEXGetPARAM_CtzResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_CtzResult.FEXErr)

    m = response.FEXGetPARAM_CtzResult.FEXResultGet
    return {'ctz': str(m.Mon_ctz), 'fecha': str(m.Mon_fecha)}


def get_param_pto_venta(client, token, sign, cuit):
    "Recuperador de los puntos de venta asignados a WSFEX"
    response = client.FEXGetPARAM_PtoVenta(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_PtoVentaResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_PtoVentaResult.FEXErr)

    pto_vtas = []
    try:
        for p in response.FEXGetPARAM_PtoVentaResult.\
                FEXResultGet.ClsFEXResponse_PtoVenta:
            pais = {
                'nro': int(p.Pve_Nro),
                'bloqueado': str(p.Pve_Bloqueado).decode('utf8'),
                'baja': str(p.Pve_FchBaja), }
            pto_vtas.append(pais)
    except AttributeError:
        pass  # no devolvi� ning�n pv (testing)
    return pto_vtas


def get_param_incoterms(client, token, sign, cuit):
    "Recuperador de valores referenciales de Incoterms"
    response = client.FEXGetPARAM_Incoterms(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetPARAM_IncotermsResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetPARAM_IncotermsResult.FEXErr)

    incoterms = []
    for p in response.FEXGetPARAM_IncotermsResult.\
            FEXResultGet.ClsFEXResponse_Inc:
        pais = {'id': str(p.Inc_Id), 'ds': str(p.Inc_Ds).decode('utf8'),
                'vig_desde': str(p.Inc_vig_desde),
                'vig_hasta': str(p.Inc_vig_hasta), }
        incoterms.append(pais)
    return incoterms


def get_last_cmp(client, token, sign, cuit, tipo_cbte, punto_vta):
    "Recuperador de ultimo valor de comprobante autorizado"
    response = client.FEXGetLast_CMP(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit),
              "Tipo_cbte": tipo_cbte,
              "Pto_venta": punto_vta})

    if int(response.FEXGetLast_CMPResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetLast_CMPResult.FEXErr)

    result = response.FEXGetLast_CMPResult.FEXResult_LastCMP
    cbte_nro = int(result.Cbte_nro)
    try:
        cbte_fecha = str(result.Cbte_fecha)
    except AttributeError:
        cbte_fecha = ""

    events = []
    for FEX_event in response.FEXGetLast_CMPResult.FEXEvents:
        code = FEX_event.EventCode
        try:
            msg = FEX_event.EventMsg
        except AttributeError:
            msg = ''
        events.append(dict(code=code, msg=msg))
    return cbte_nro, cbte_fecha, events


def get_last_id(client, token, sign, cuit):
    "Recupera el ultimo ID y su fecha"
    response = client.FEXGetLast_ID(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.FEXGetLast_IDResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetLast_IDResult.FEXErr)

    result = response.FEXGetLast_IDResult.FEXResultGet
    id = int(result.Id)
    events = []
    for FEX_event in response.FEXGetLast_IDResult.FEXEvents:
        code = FEX_event.EventCode
        try:
            msg = FEX_event.EventMsg
        except AttributeError:
            msg = ''
        events.append(dict(code=code, msg=msg))
    return id, events


def authorize(client, token, sign, cuit, id, factura):
    "Autorizador: recibe la factura como un dict, devuelve auth y events"

    # Informaci�n de la autenticaci�n
    auth = {"Token": token, "Sign": sign, "Cuit": long(cuit)}
    # Informaci�n de la factura de ingreso
    cmp = {"Id": id}
    cmp.update(factura)

    # llamo al webservice para autorizar la factura
    response = client.FEXAuthorize(Auth=auth, Cmp=cmp)

    # hubo error?
    if int(response.FEXAuthorizeResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXAuthorizeResult.FEXErr)

    # extraigo la respuesta (auth y eventos)
    result = response.FEXAuthorizeResult.FEXResultAuth
    auth = dict(
        id=int(result.Id), cuit=int(result.Cuit), cae=str(result.Cae),
        fch_cbte=str(result.Fch_cbte), fch_venc_cae=str(result.Fch_venc_Cae),
        resultado=str(result.Resultado), cbte_nro=str(result.Cbte_nro),
        reproceso=str(result.Reproceso), obs=str(result.Motivos_Obs))
    events = []
    for FEX_event in response.FEXAuthorizeResult.FEXEvents:
        code = FEX_event.EventCode
        try:
            msg = FEX_event.EventMsg
        except AttributeError:
            msg = ''
        events.append(dict(code=code, msg=msg))

    return auth, events


def get_cmp(client, token, sign, cuit, tipo_cbte, punto_vta, cbte_nro):
    "Recupera los datos completos de un comprobante ya autorizado"
    response = client.FEXGetCMP(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)},
        Cmp={"Tipo_cbte": tipo_cbte,
             "Punto_vta": punto_vta, "Cbte_nro": cbte_nro})

    if int(response.FEXGetCMPResult.FEXErr.ErrCode) != 0:
        raise FEXError(response.FEXGetCMPResult.FEXErr)

    result = response.FEXGetCMPResult.FEXResultGet
    # print(result.Id)
    cbt = dict(cuit=int(result.Cuit_pais_cliente), cae=str(result.Cae),
               fch_cbte=str(result.Fecha_cbte),
               fch_venc_cae=str(result.Fch_venc_Cae),
               imp_total=str(result.Imp_total),
               obs=str(result.Obs))
    events = []
    for FEX_event in response.FEXGetCMPResult.FEXEvents:
        code = FEX_event.EventCode
        try:
            msg = FEX_event.EventMsg
        except AttributeError:
            msg = ''
        events.append(dict(code=code, msg=msg))
    return cbt, events


# Clases para facilitar la autorizaci�n de facturas
class FacturaEX:
    "Factura Electr�nica Exportaci�n"
    # valores por defecto del encabezado de la factura
    tipo_cbte = 19
    punto_vta = 1
    cbte_nro = 0
    fecha_cbte = None
    imp_total = 0.0
    tipo_expo = 1
    permiso_existente = 1
    dst_cmp = None
    cliente = cuit_pais_cliente = domicilio_cliente = id_impositivo = ""
    moneda_id = "PES"
    moneda_ctz = 1.0
    obs_comerciales = obs = forma_pago = incoterms = ""
    idioma_cbte = 7
    items = None

    def __init__(self, **kwargs):
        self.items = []
        self.permisos = []
        self.cmps_asoc = []
        for k, v in kwargs.items():
            setattr(self, k, v)

    def add_item(self, item, calcular=True):
        self.items.append(item)
        if calcular:
            self.imp_total += item.imp_total

    def add_permiso(self, permiso):
        self.permisos.append(permiso)

    def add_cmp_asoc(self, cmp_asoc):
        self.cmps_asoc.append(cmp_asoc)

    def to_dict(self, completo=False):
        "Convierte el objeto factura en un diccionario serializable por xml"
        dic = {
            'Fecha_cbte': self.fecha_cbte,
            'Tipo_cbte': self.tipo_cbte,
            'Punto_vta': self.punto_vta,
            'Cbte_nro': self.cbte_nro,
            'Tipo_expo': self.tipo_expo,
            'Permiso_existente': self.permiso_existente,
            'Dst_cmp': self.dst_cmp,
            'Cliente': self.cliente,
            'Cuit_pais_cliente': self.cuit_pais_cliente,
            'Domicilio_cliente': self.domicilio_cliente,
            'Id_impositivo': self.id_impositivo,
            'Moneda_Id': self.moneda_id,
            'Moneda_ctz': self.moneda_ctz,
            'Obs_comerciales': self.obs_comerciales,
            'Imp_total': self.imp_total,
            'Obs': self.obs,
            'Forma_pago': self.forma_pago,
            'Incoterms': self.incoterms,
            'Idioma_cbte': self.idioma_cbte,
            'Items': [item.to_dict() for item in self.items],
            }
        if self.permisos:
            dic['Permisos'] = [permiso.to_dict() for permiso in self.permisos]
        if self.cmps_asoc:
            dic['Cmps_asoc'] = [cmp.to_dict() for cmp in self.cmps_asoc]
        return dic


class ItemFEX:
    "Item de Factura Electr�nica Exportaci�n"
    # valores por defecto:
    pro_codigo = ""
    pro_ds = ""
    pro_qty = 0.0
    pro_umed = 1
    pro_precio_uni = 0.0
    imp_total = 0.0

    def __init__(
        self, codigo="", ds="", qty=1, umed=7, precio=0.00,
        imp_total=0.00, **kwargs
    ):
        "Constructor"
        self.pro_codigo = codigo
        self.pro_ds = ds
        self.pro_qty = qty
        self.pro_umed = umed
        self.pro_precio_uni = precio
        self.imp_total = imp_total or (precio*qty)

    def to_dict(self):
        "Convierte el objeto item en un diccionario serializable por xml"
        return {
           'Item': {
             'Pro_codigo': self.pro_codigo,
             'Pro_ds': self.pro_ds,
             'Pro_qty': self.pro_qty,
             'Pro_umed': self.pro_umed,
             'Pro_precio_uni': self.pro_precio_uni,
             'Pro_total_item': self.imp_total}
           }


class PermisoFEX:
    "Permiso de Factura Electr�nica Exportaci�n"
    # valores por defecto:
    id_permiso = ""
    dst_merc = 0

    def __init__(self, id_permiso="", dst_merc="",  **kwargs):
        "Constructor"
        self.id_permiso = id_permiso
        self.dst_merc = dst_merc

    def to_dict(self):
        "Convierte el objeto permiso en un diccionario serializable por xml"
        return {
           'Permiso': {
             'Id_permiso': self.id_permiso,
             'Dst_merc': self.dst_merc}
           }


class CmpAsocFEX:
    "Comprobante Asociado de Factura Electr�nica Exportaci�n"
    # valores por defecto:
    tipo = 0
    punto_vta = 0
    nro = 0

    def __init__(self, cbte_tipo=0, cbte_punto_vta=0, cbte_nro=0, **kwargs):
        "Constructor"
        self.tipo = cbte_tipo
        self.punto_vta = cbte_punto_vta
        self.nro = cbte_nro

    def to_dict(self):
        "Convierte el objeto permiso en un diccionario serializable por xml"
        return {
           'Cmp_asoc': {
             'CBte_tipo': self.tipo,
             'Cbte_punto_vta': self.punto_vta,
             'Cbte_nro': self.nro}
           }


def main():
    "Funci�n principal de pruebas (obtener CAE)"
    import os
    import time

    url = len(sys.argv) > 2 and sys.argv[2].startswith("http") and \
        sys.argv[2] or WSFEXURL

    # cliente soap del web service
    client = SoapClient(
        url,
        action=SOAP_ACTION,
        namespace=SOAP_NS,
        trace="--trace" in sys.argv)

    # obteniendo el TA
    TA = "TA.xml"
    if not os.path.exists(TA) or os.path.getmtime(TA)+(60*60*5) < time.time():
        import wsaa
        tra = wsaa.create_tra(service="wsfex")
        cms = wsaa.sign_tra(tra, "reingart.crt", "reingart.key")
        ta_string = wsaa.call_wsaa(cms)
        open("TA.xml", "w").write(ta_string)
    ta_string = open(TA).read()
    ta = SimpleXMLElement(ta_string)
    token = str(ta.credentials.token)
    sign = str(ta.credentials.sign)
    # fin TA

    CUIT = len(sys.argv) > 1 and sys.argv[1] or "20267565393"

    if "--dummy" in sys.argv:
        print(dummy(client))

    # Recuperar par�metros:
    if "--params" in sys.argv:
        import codecs
        # import locale
        sys.stdout = codecs.getwriter('latin1')(sys.stdout)

        print("=== Monedas ===")
        monedas = get_param_mon(client, token, sign, CUIT)
        for m in monedas:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % m)
        monedas = dict([(m['id'], m['ds']) for m in monedas])

        print("=== Tipos Comprobante ===")
        cbtes = get_param_tipo_cbte(client, token, sign, CUIT)
        for c in cbtes:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % c)
        comprobantes = dict([(c['id'], c['ds']) for c in cbtes])

        print(u"=== Tipos Exportaci�n ===")
        tipos = get_param_tipo_expo(client, token, sign, CUIT)
        for t in tipos:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % t)
        tipos_expo = dict([(t['id'], t['ds']) for t in tipos])

        print("=== Idiomas ===")
        tipos = get_param_idiomas(client, token, sign, CUIT)
        for t in tipos:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % t)
        idiomas = dict([(t['id'], t['ds']) for t in tipos])

        print("=== Unidades de medida ===")
        umedidas = get_param_umed(client, token, sign, CUIT)
        for u in umedidas:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % u)
        umeds = dict([(u['id'], u['ds']) for u in umedidas])

        print("=== INCOTERMs ===")
        incoterms = get_param_incoterms(client, token, sign, CUIT)
        for i in incoterms:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % i)
        incoterms = dict([(t['id'], t['ds']) for t in incoterms])

        print("=== Pais Destino ===")
        pais = get_param_dst_pais(client, token, sign, CUIT)
        for p in pais:
            print("||%(codigo)s||%(ds)s||" % p)
        paises = dict([(p['codigo'], p['ds']) for p in pais])
        print(paises)

        print("=== CUIT Pais Destino ===")
        pais = get_param_dst_cuit(client, token, sign, CUIT)
        for p in pais:
            print("||%(cuit)s||%(ds)s||" % p)
        cuits = dict([(p['cuit'], p['ds']) for p in pais])
        print(cuits)

        ctz = get_param_ctz(client, token, sign, CUIT, 'DOL')
        print(ctz)

        print(get_param_pto_venta(client, token, sign, CUIT))

    else:
        monedas = {
            'DOL': u'D\xf3lar Estadounidense', 'PES': u'Pesos Argentinos',
            '010': u'Pesos Mejicanos', '011': u'Pesos Uruguayos',
            '012': u'Real', '014': u'Coronas Danesas',
            '015': u'Coronas Noruegas', '016': u'Coronas Suecas',
            '019': u'Yens', '018': u'D\xf3lar Canadiense',
            '033': u'Peso Chileno', '056': u'Forint (Hungr\xeda)',
            '031': u'Peso Boliviano', '036': u'Sucre Ecuatoriano',
            '051': u'D\xf3lar de Hong Kong', '034': u'Rand Sudafricano',
            '053': u'D\xf3lar de Jamaica', '057': u'Baht (Tailandia)',
            '043': u'Balboas Paname\xf1as', '042': u'Peso Dominicano',
            '052': u'D\xf3lar de Singapur', '032': u'Peso Colombiano',
            '035': u'Nuevo Sol Peruano', '061': u'Zloty Polaco',
            '060': u'Euro', '063': u'Lempira Hondure\xf1a',
            '062': u'Rupia Hind\xfa', '064': u'Yuan (Rep. Pop. China)',
            '009': u'Franco Suizo', '025': u'Dinar Yugoslavo',
            '002': u'D\xf3lar Libre EEUU', '027': u'Dracma Griego',
            '026': u'D\xf3lar Australiano', '007': u'Florines Holandeses',
            '023': u'Bol\xedvar Venezolano', '047': u'Riyal Saudita',
            '046': u'Libra Egipcia', '045': u'Dirham Marroqu\xed',
            '044': u'C\xf3rdoba Nicarag\xfcense', '029': u'G\xfcaran\xed',
            '028': u'Flor\xedn (Antillas Holandesas)',
            '054': u'D\xf3lar de Taiwan', '040': u'Lei Rumano',
            '024': u'Corona Checa', '030': u'Shekel (Israel)',
            '021': u'Libra Esterlina', '055': u'Quetzal Guatemalteco',
            '059': u'Dinar Kuwaiti'}
        comprobantes = {
            19: u'Facturas de Exportaci\xf3n\n',
            20: u'Nota de D\xe9bito por Operaciones con el Exterior\n',
            21: u'Nota de Cr\xe9dito por Operaciones con el Exterior\n'}
        print(comprobantes)
        tipos_expo = {
            1: u'Exportaci\xf3n definitiva de Bienes',
            2: u'Servicios',
            4: u'Otros'}
        print(tipos_expo)
        idiomas = {1: u'Espa\xf1ol', 2: u'Ingl\xe9s', 3: u'Portugu\xe9s'}
        print(idiomas)
        umeds = {
            0: u' ', 1: u'kilogramos', 2: u'metros',
            3: u'metros cuadrados', 4: u'metros c\xfabicos', 5: u'litros',
            6: u'1000 kWh', 7: u'unidades', 8: u'pares',
            9: u'docenas', 10: u'quilates', 11: u'millares',
            14: u'gramos', 15: u'milimetros', 16: u'mm c\xfabicos',
            17: u'kil\xf3metros', 18: u'hectolitros', 20: u'cent\xedmetros',
            25: u'jgo. pqt. mazo naipes', 27: u'cm c\xfabicos',
            29: u'toneladas',
            30: u'dam c\xfabicos', 31: u'hm c\xfabicos', 32: u'km c\xfabicos',
            33: u'microgramos', 34: u'nanogramos', 35: u'picogramos',
            41: u'miligramos', 47: u'mililitros', 48: u'curie',
            49: u'milicurie', 50: u'microcurie', 51: u'uiacthor',
            52: u'muiacthor', 53: u'kg base', 54: u'gruesa',
            61: u'kg bruto', 62: u'uiactant', 63: u'muiactant',
            64: u'uiactig', 65: u'muiactig', 66: u'kg activo',
            67: u'gramo activo', 68: u'gramo base', 96: u'packs',
            97: u'hormas', 98: u'bonificaci\xf3n', 99: u'otras unidades'}
        print(umeds)
        incoterms = {
            'DAF': u'DAF', 'DDP': u'DDP', 'CIF': u'CIF',
            'FCA': u'FCA', 'FAS': u'FAS', 'DES': u'DES',
            'CPT': u'CPT', 'EXW': u'EXW', 'CIP': u'CIP',
            'DDU': u'DDU', 'FOB': u'FOB', 'DEQ': u'DEQ', 'CFR': u'CFR'}

    if "--prueba" in sys.argv:

        # recupero ultimo comprobante y id
        tipo_cbte = 19
        punto_vta = 1
        cbte_nro, cbte_fecha, events = get_last_cmp(
            client, token, sign, CUIT, tipo_cbte, punto_vta)
        id, events = get_last_id(client, token, sign, CUIT)

        print("last_cmp", cbte_nro, cbte_fecha)
        print("last_id", id)

        cbte_nro += 1
        id += 1

        if False:  # prueba a mano
            f = {
                'Id': id,
                'Fecha_cbte': date('Ymd'),
                'Tipo_cbte': tipo_cbte,
                'Punto_vta': punto_vta,
                'Cbte_nro': cbte_nro,
                'Tipo_expo': 1,
                'Permiso_existente': 'N',
                'Dst_cmp': 225,
                'Cliente': 'Jose Yorugua',
                'Cuit_pais_cliente': 50000000016,
                'Domicilio_cliente': 'Montevideo, UY',
                'Id_impositivo': 'RUC 123123',
                'Moneda_Id': 'DOL',
                'Moneda_ctz': 3.85,
                'Obs_comerciales': 'Observaciones comerciales',
                'Imp_total': 60.00,
                'Obs': 'Observaciones',
                'Forma_pago': 'Taka taka',
                'Incoterms': 'FOB',
                'Idioma_cbte': 2,
                'Items': [{
                   'Item': {
                     'Pro_codigo': "kbd",
                     'Pro_ds': "Keyboard (uruguayan layout)",
                     'Pro_qty': 1,
                     'Pro_umed': 7,
                     'Pro_precio_uni': 50.000049,
                     'Pro_total_item': 50.000049}
                   }, {
                   'Item': {
                     'Pro_codigo': "mice",
                     'Pro_ds': "Optical Mouse",
                     'Pro_qty': 1,
                     'Pro_umed': 7,
                     'Pro_precio_uni': 10.50,
                     'Pro_total_item': 10.50}
                   }]
                }
        else:
            # creo una factura de prueba
            f = FacturaEX()
            f.punto_vta = punto_vta
            f.cbte_nro = cbte_nro
            f.fecha_cbte = date('Ymd')
            f.tipo_expo = 1
            f.permiso_existente = 'S'
            f.dst_cmp = 203
            f.cliente = 'Joao Da Silva'
            f.cuit_pais_cliente = 50000000016
            f.domicilio_cliente = 'Rua 76 km 34.5 Alagoas'
            f.id_impositivo = 'PJ54482221-l'
            f.moneda_id = '012'
            f.moneda_ctz = 0.5
            f.obs_comerciales = 'Observaciones comerciales'
            f.obs = 'Sin observaciones'
            f.forma_pago = '30 dias'
            f.incoterms = 'FOB'
            f.idioma_cbte = 1
            # agrego los items
            it = ItemFEX(
                codigo='PRO1',
                ds=u'Producto Tipo 1 Exportacion MERCOSUR ISO 9001',
                qty=1, precio=125)
            f.add_item(it)
            it = ItemFEX(
                codigo='PRO2',
                ds=u'Producto Tipo 2 Exportacion MERCOSUR ISO 9001',
                qty=1, precio=125)
            f.add_item(it)
            permiso = PermisoFEX(id_permiso="99999AAXX999999A", dst_merc=225)
            f.add_permiso(permiso)
            permiso = PermisoFEX(id_permiso="99999AAXX999999B", dst_merc=225)
            f.add_permiso(permiso)
            if f.tipo_cbte != 19:
                cmp_asoc = CmpAsocFEX(tipo=19, punto_vta=2, cbte_nro=1)
                f.add_cmp_asoc(cmp_asoc)
                cmp_asoc = CmpAsocFEX(tipo=19, punto_vta=2, cbte_nro=2)
                f.add_cmp_asoc(cmp_asoc)
            f = f.to_dict()
            print(f)

        try:
            auth, events = authorize(
                client, token, sign, CUIT, id=id, factura=f)
        except FEXError as fex:
            print("FEX_error:", fex.msg)
            print(client.xml_response)
            sys.exit(1)
        cae = auth['cae']
        print("cae", cae)
        print("auth", auth)
        print("events", events)
        try:
            auth, events = get_cmp(
                client, token, sign, CUIT, tipo_cbte, punto_vta, cbte_nro)
        except Exception as e:
            print(e)
            print(client.xml_response)
            sys.exit(1)
        print("get_cmp: auth", auth)
        print("get_cmp: events", events)

    if '--ult' in sys.argv:
        print("Consultar ultimo numero:")
        tipo_cbte = int(raw_input("Tipo de comprobante: "))
        punto_vta = int(raw_input("Punto de venta: "))
        try:
            ult_cbte, fecha, events = get_last_cmp(
                client, token, sign, CUIT, tipo_cbte, punto_vta)
            print("Ultimo numero: ", ult_cbte)
            print("Fecha: ", fecha)
        finally:
            print(client.xml_request)
            print(client.xml_response)

    if '--get' in sys.argv:
        print("Recuperar comprobante:")
        tipo_cbte = int(raw_input("Tipo de comprobante: "))
        punto_vta = int(raw_input("Punto de venta: "))
        cbte_nro = int(raw_input("Numero de comprobante: "))
        try:
            cbt, events = get_cmp(
                client, token, sign, CUIT, tipo_cbte, punto_vta, cbte_nro)
            for k, v in cbt.items():
                print("%s = %s" % (k, v))
        finally:
            print(client.xml_request)
            print(client.xml_response)
    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para obtener código de autorización electrónica (CAE) para
Liquidación de Tabaco Verde del web service WSLTV de AFIP
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2016 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.01a"

LICENCIA = """
wsltv.py: Interfaz para generar Código de Autorización Electrónica (CAE) para
Liquidación de Tabaco Verde (LtvService)
Copyright (C) 2016 Mariano Reingart reingart@gmail.com
http://www.sistemasagiles.com.ar/trac/wiki/LiquidacionPrimariaGranos

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo respetando la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

AYUDA="""
Opciones:
  --ayuda: este mensaje

  --debug: modo depuración (detalla y confirma las operaciones)
  --formato: muestra el formato de los archivos de entrada/salida
  --prueba: genera y autoriza una liquidación de prueba (no usar en producción!)
  --xml: almacena los requerimientos y respuestas XML (depuración)
  --dbf: utilizar tablas DBF (xBase) para los archivos de intercambio
  --json: utilizar formato json para el archivo de intercambio
  --dummy: consulta estado de servidores

  --autorizar: Autorizar Liquidación de Tabaco Verde (generarLiquidacion)
  --ajustar: Ajustar Liquidación de Tabaco Verde (ajustarLiquidacion)
  --ult: Consulta el último número de orden registrado en AFIP
         (consultarUltimoComprobanteXPuntoVenta)
  --consultar: Consulta una liquidación registrada en AFIP
         (consultarLiquidacionXNroComprobante / consultarLiquidacionXCAE)

  --pdf: descarga la liquidación en formato PDF
  --mostrar: muestra el documento PDF generado (usar con --pdf)
  --imprimir: imprime el documento PDF generado (usar con --mostrar y --pdf)

  --provincias: obtiene el listado de provincias (código/descripción)
  --condicionesventa: obtiene el listado de las condiciones de venta
  --tributos: obtiene el listado de los tributos
  --retenciones: obtiene el listado de las retenciones de tabaco
  --variedades: obtiene el listado de las variedades y especies de tabaco
  --depositos: obtiene el listado de los depositos de acopio (para el contrib.)
  --puntosventa: obtiene el listado de puntos de venta habilitados

Ver wsltv.ini para parámetros de configuración (URL, certificados, etc.)"
"""

import os, sys, shelve
import decimal, datetime
import traceback
import pprint
from py3simplesoap.client import SoapFault
from fpdf import Template
from . import utils

# importo funciones compartidas:
from .utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I, json, BaseWS, inicializar_y_capturar_excepciones, get_install_dir


WSDL = "https://fwshomo.afip.gov.ar/wsltv/LtvService?wsdl"
#WSDL = "https://serviciosjava.afip.gob.ar/wsltv/LtvService?wsdl"

DEBUG = False
XML = False
CONFIG_FILE = "wsltv.ini"
HOMO = True


class WSLTV(BaseWS):
    "Interfaz para el WebService de Liquidación de Tabaco Verde"
    _public_methods_ = ['Conectar', 'Dummy', 'SetTicketAcceso', 'DebugLog',
                        'AutorizarLiquidacion',
                        'CrearLiquidacion',
                        'AgregarCondicionVenta', 'AgregarReceptor',
                        'AgregarRomaneo', 'AgregarFardo', 'AgregarPrecioClase',
                        'AgregarRetencion', 'AgregarTributo',
                        'ConsultarLiquidacion', 'ConsultarUltimoComprobante',
                        'CrearAjuste', 'AgregarComprobanteAAjustar',
                        'AjustarLiquidacion',
                        'LeerDatosLiquidacion',
                        'ConsultarVariedadesClasesTabaco',
                        'ConsultarTipoTributos',
                        'ConsultarRetencionesTabacaleras',
                        'ConsultarDepositosAcopio',
                        'ConsultarPuntosVentas',
                        'ConsultarProvincias',
                        'ConsultarCondicionesVenta',
                        'MostrarPDF',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        'SetParametros', 'SetParametro', 'GetParametro',
                        ]
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'Excepcion', 'ErrCode', 'ErrMsg', 'LanzarExcepciones', 'Errores',
        'XmlRequest', 'XmlResponse', 'Version', 'Traceback', 'InstallDir',
        'CAE', 'NroComprobante', 'FechaLiquidacion',
        'ImporteNeto', 'TotalRetenciones', 'TotalTributos', 'Total',
        'AlicuotaIVA', 'ImporteIVA', 'Subtotal',
        ]
    _reg_progid_ = "WSLTV"
    _reg_clsid_ = "{C6EEAE8A-7560-4538-B29C-76434A8C2DC3}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    LanzarExcepciones = False
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.errores = []
        self.CAE = ""
        self.NroComprobante = self.FechaLiquidacion = ''
        self.ImporteNeto = ""
        self.ImporteIVA = ""
        self.AlicuotaIVA = ""
        self.TotalRetenciones = ""
        self.TotalTributos = ""
        self.Subtotal = self.Total = ""
        self.datos = {}

    @inicializar_y_capturar_excepciones
    def Conectar(self, cache=None, url="", proxy="", wrapper="", cacert=None, timeout=30):
        "Establecer la conexión a los servidores de la AFIP"
        # llamo al constructor heredado:
        ok = BaseWS.Conectar(self, cache, url, proxy, wrapper, cacert, timeout)
        if ok:
            # corrijo ubicación del servidor (puerto htttp 80 en el WSDL)
            location = self.client.services['LtvService']['ports']['LtvEndPoint']['location']
            if location.startswith("http://"):
                print("Corrigiendo WSDL ...", location, end=' ')
                location = location.replace("http://", "https://").replace(":80", ":443")
                self.client.services['LtvService']['ports']['LtvEndPoint']['location'] = location
                print(location)
        return ok

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        errores = []
        if 'errores' in ret:
            errores.extend(ret['errores'])
        if errores:
            self.Errores = ["%(codigo)s: %(descripcion)s" % err['error'][0]
                            for err in errores]
            self.errores = [
                {'codigo': err['error'][0]['codigo'],
                 'descripcion': err['error'][0]['descripcion'].replace("\n", "")
                                .replace("\r", "")}
                             for err in errores]
            self.ErrCode = ' '.join(self.Errores)
            self.ErrMsg = '\n'.join(self.Errores)

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.dummy()['respuesta']
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])
        return True

    @inicializar_y_capturar_excepciones
    def CrearLiquidacion(self, tipo_cbte, pto_vta, nro_cbte, fecha,
            cod_deposito_acopio, tipo_compra,
            variedad_tabaco, cod_provincia_origen_tabaco,
            puerta=None, nro_tarjeta=None, horas=None, control=None,
            nro_interno=None, iibb_emisor=None,
            **kwargs):
        "Inicializa internamente los datos de una liquidación para autorizar"
        # creo el diccionario con los campos generales de la liquidación:
        liq = dict(tipoComprobante=tipo_cbte,
                   nroComprobante=nro_cbte,
                   puntoVenta=pto_vta,
                   iibbEmisor=iibb_emisor,
                   codDepositoAcopio=cod_deposito_acopio,
                   fechaLiquidacion=fecha,
                   tipoCompra=tipo_compra,
                   condicionVenta=[],
                   variedadTabaco=variedad_tabaco,
                   codProvinciaOrigenTabaco=cod_provincia_origen_tabaco,
                   puerta=puerta,
                   nroTarjeta=nro_tarjeta,
                   horas=horas,
                   control=control,
                   nroInterno=nro_interno,
                   )
        self.solicitud = dict(liquidacion=liq,
                              receptor={},
                              romaneo=[],
                              precioClase=[],
                              retencion=[],
                              tributo=[],
                             )
        return True

    @inicializar_y_capturar_excepciones
    def AgregarCondicionVenta(self, codigo, descripcion, **kwargs):
        "Agrego una o más condicion de venta a la liq."
        cond = {'codigo': codigo, 'descripcion': descripcion}
        self.solicitud['liquidacion']['condicionVenta'].append(cond)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarReceptor(self, cuit, iibb, nro_socio, nro_fet, **kwargs):
        "Agrego un receptor a la liq."
        rcpt = dict(cuit=cuit, iibb=iibb, nroSocio=nro_socio, nroFET=nro_fet)
        self.solicitud['receptor'] = rcpt
        return True

    @inicializar_y_capturar_excepciones
    def AgregarRomaneo(self, nro_romaneo, fecha_romaneo, **kwargs):
        "Agrego uno o más romaneos a la liq."
        romaneo = dict(nroRomaneo=nro_romaneo, fechaRomaneo=fecha_romaneo,
                       fardo=[])
        self.solicitud['romaneo'].append(romaneo)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarFardo(self, cod_trazabilidad, clase_tabaco, peso, **kwargs):
        "Agrego un fardo al último romaneo agregado a la liq."
        fardo = dict(codTrazabilidad=cod_trazabilidad, claseTabaco=clase_tabaco, peso=peso)
        self.solicitud['romaneo'][-1]['fardo'].append(fardo)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarPrecioClase(self, clase_tabaco, precio, **kwargs):
        "Agrego un PrecioClase a la liq."
        precioclase = dict(claseTabaco=clase_tabaco, precio=precio,)
        self.solicitud['precioClase'].append(precioclase)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarRetencion(self, cod_retencion, descripcion, importe, **kwargs):
        "Agrega la información referente a las retenciones de la liquidación"
        ret = dict(codRetencion=cod_retencion, descripcion=descripcion, importe=importe)
        self.solicitud['retencion'].append(ret)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarTributo(self, codigo_tributo, descripcion, base_imponible, alicuota, importe):
        "Agrega la información referente a las retenciones de la liquidación"
        trib = dict(codigoTributo=codigo_tributo, descripcion=descripcion, baseImponible=base_imponible, alicuota=alicuota, importe=importe)
        self.solicitud['tributo'].append(trib)
        return True


    @inicializar_y_capturar_excepciones
    def AutorizarLiquidacion(self):
        "Autorizar Liquidación Electrónica de Tabaco Verde"
        # llamo al webservice:
        ret = self.client.generarLiquidacion(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                        solicitud=self.solicitud,
                        )
        # analizo la respusta
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.AnalizarLiquidacion(ret.get('liquidacion'))
        return True


    def AnalizarLiquidacion(self, liq):
        "Método interno para analizar la respuesta de AFIP"
        # proceso los datos básicos de la liquidación (devuelto por consultar):
        if liq:
            cab = liq['cabecera']
            self.CAE = str(cab['cae'])
            self.FechaLiquidacion = cab['fechaLiquidacion']
            self.NroComprobante = cab['nroComprobante']
            tot = liq['totalesOperacion']
            self.AlicuotaIVA = tot['alicuotaIVA']
            self.ImporteNeto = tot['importeNeto']
            self.ImporteIVA = tot['importeIVA']
            self.Subtotal = tot['subtotal']
            self.TotalRetenciones = tot['totalRetenciones']
            self.TotalTributos = tot['totalTributos']
            self.Total = tot['total']

            # parámetros de salida:
            self.params_out = dict(
                tipo_cbte=liq['cabecera']['tipoComprobante'],
                pto_vta=liq['cabecera']['puntoVenta'],
                nro_cbte=liq['cabecera']['nroComprobante'],
                fecha=liq['cabecera']['fechaLiquidacion'],
                cod_deposito_acopio=liq['cabecera']['codDepositoAcopio'],
                cae=str(liq['cabecera']['cae']),
                domicilio_punto_venta=liq['cabecera']['domicilioPuntoVenta'],
                domicilio_deposito_acopio=liq['cabecera']['domicilioDepositoAcopio'],
                emisor=dict(
                    cuit=liq['emisor']['cuit'],
                    razon_social=liq['emisor']['razonSocial'],
                    situacion_iva=liq['emisor']['situacionIVA'],
                    domicilio=liq['emisor']['domicilio'],
                    fecha_inicio_actividad=liq['emisor']['fechaInicioActividad'],
                    ),
                receptor=dict(
                    cuit=liq['receptor']['cuit'],
                    razon_social=liq['receptor']['razonSocial'],
                    nro_fet=liq['receptor']['nroFET'],
                    nro_socio=liq['receptor']['nroSocio'],
                    situacion_iva=liq['receptor']['situacionIVA'],
                    domicilio=liq['receptor']['domicilio'],
                    iibb=liq['receptor']['iibb'],
                    ),
                control=liq['datosOperacion']['control'],
                nro_interno=liq['datosOperacion']['nroInterno'],
                condicion_venta=liq['datosOperacion']['condicionVenta'],
                variedad_tabaco=liq['datosOperacion']['variedadTabaco'],
                puerta=liq['datosOperacion']['puerta'],
                nro_tarjeta=liq['datosOperacion']['nroTarjeta'],
                horas=liq['datosOperacion']['horas'],
                cod_provincia_origen_tabaco=liq['datosOperacion']['codProvinciaOrigenTabaco'],
                tipo_compra=liq['datosOperacion']['tipoCompra'],
                peso_total_fardos_kg=liq['detalleOperacion']['pesoTotalFardosKg'],
                cantidad_total_fardos=liq['detalleOperacion']['cantidadTotalFardos'],
                romaneos=[],
                alicuota_iva=liq['totalesOperacion']['alicuotaIVA'],
                importe_iva=liq['totalesOperacion']['importeIVA'],
                importe_neto=liq['totalesOperacion']['importeNeto'],
                subtotal=liq['totalesOperacion']['subtotal'],
                total_retenciones=liq['totalesOperacion']['totalRetenciones'],
                total_tributos=liq['totalesOperacion']['totalTributos'],
                total=liq['totalesOperacion']['total'],
                retenciones=[],
                tributos=[],
                pdf=liq.get('pdf'),
                )
            for romaneo in liq['detalleOperacion']['romaneo']:
                self.params_out['romaneos'].append(dict(
                    fecha_romaneo=romaneo['fechaRomaneo'],
                    nro_romaneo=romaneo['nroRomaneo'],
                    detalle_clase=[dict(
                        cantidad_fardos=det['cantidadFardos'],
                        cod_clase=det['codClase'],
                        importe=det['importe'],
                        peso_fardos_kg=det['pesoFardosKg'],
                        precio_x_kg_fardo=det['precioXKgFardo'],
                        ) for det in romaneo['detalleClase']],
                    ))
            for ret in liq['retencion']:
                self.params_out['retenciones'].append(dict(
                    retencion_codigo=ret['codigo'],
                    retencion_importe=ret['importe'],
                    ))
            for trib in liq['tributo']:
                self.params_out['tributos'].append(dict(
                    tributo_descripcion=trib['descripcion'],
                    tributo_base_imponible=trib['baseImponible'],
                    tributo_alicuota=trib['alicuota'],
                    tributo_codigo=trib['codigo'],
                    tributo_importe=trib['importe'],
                    ))
            if DEBUG:
                import pprint
                pprint.pprint(self.params_out)
        self.params_out['errores'] = self.errores

    @inicializar_y_capturar_excepciones
    def CrearAjuste(self, tipo_cbte, pto_vta, nro_cbte, fecha,
            cod_deposito_acopio, tipo_ajuste, cuit_receptor,
            iibb_emisor=None, iibb_receptor=None,
            **kwargs):
        "Inicializa internamente los datos de una liquidación para ajustar"
        # creo el diccionario con los campos generales de la liquidación:
        liq = dict(tipoComprobante=tipo_cbte,
                   nroComprobante=nro_cbte,
                   puntoVenta=pto_vta,
                   fechaAjusteLiquidacion=fecha,
                   codDepositoAcopio=cod_deposito_acopio,
                   tipoAjuste=tipo_ajuste,
                   cuitReceptor=cuit_receptor,
                   iibbReceptor=iibb_receptor,
                   iibbEmisor=iibb_emisor,
                   comprobanteAAjustar=[],
                   )
        self.solicitud = dict(liquidacion=liq,
                              receptor=[],
                              romaneo=[],
                              precioClase=[],
                              retencion=[],
                              tributo=[],
                             )
        return True

    @inicializar_y_capturar_excepciones
    def AgregarComprobanteAAjustar(self, tipo_cbte, pto_vta, nro_cbte):
        "Agrega comprobante a ajustar"
        cbte = dict(tipoComprobante=tipo_cbte, puntoVenta=pto_vta, nroComprobante=nro_cbte)
        self.solicitud['liquidacion'].append(cbte)
        return True

    @inicializar_y_capturar_excepciones
    def AjustarLiquidacion(self):
        "Ajustar Liquidación de Tabaco Verde"

        # renombrar la clave principal de la estructura
        liq = self.solicitud.pop('liquidacion')
        self.solicitud["liquidacionAjuste"] = liq

        # llamar al webservice:
        ret = self.client.ajustarLiquidacion(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                        solicitud=self.solicitud,
                        )
        # analizar el resultado:
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        if 'ajusteUnificado' in ret:
            aut = ret['ajusteUnificado']
            self.AnalizarAjuste(aut)
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarLiquidacion(self, pto_vta=None, nro_cbte=None, cae=None,
                                   pdf=True):
        "Consulta una liquidación por No de Comprobante o CAE"
        if cae:
            ret = self.client.consultarLiquidacionXCAE(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                        solicitud={
                            'cae': cae,
                            'pdf': pdf,
                            },
                        )
        else:
            ret = self.client.consultarLiquidacionXNroComprobante(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                        solicitud={
                            'puntoVenta': pto_vta,
                            'nroComprobante': nro_cbte,
                            'tipoComprobante': tipo_cbte,
                            'pdf': pdf,
                            },
                        )
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        if 'liquidacion' in ret:
            liq = ret['liquidacion']
            self.AnalizarLiquidacion(liq)
        # guardo el PDF si se indico archivo y vino en la respuesta:
        if pdf and 'pdf' in ret:
            open(pdf, "wb").write(ret['pdf'])
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarUltimoComprobante(self, tipo_cbte=151, pto_vta=1):
        "Consulta el último No de Comprobante registrado"
        ret = self.client.consultarUltimoComprobanteXPuntoVenta(
                    auth={
                        'token': self.Token, 'sign': self.Sign,
                        'cuit': self.Cuit, },
                    solicitud={
                        'puntoVenta': pto_vta,
                        'tipoComprobante': tipo_cbte},
                    )
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.NroComprobante = ret['nroComprobante']
        return True


    def ConsultarProvincias(self, sep="||"):
        "Consulta las provincias habilitadas"
        ret = self.client.consultarProvincias(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('provincia', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarCondicionesVenta(self, sep="||"):
        "Retorna un listado de códigos y descripciones de las condiciones de ventas"
        ret = self.client.consultarCondicionesVenta(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('condicionVenta', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarTributos(self, sep="||"):
        "Retorna un listado de tributos con código, descripción y signo."
        ret = self.client.consultarTributos(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('tributo', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarVariedadesClasesTabaco(self, sep="||"):
        "Retorna un listado de variedades y clases de tabaco"
        ret = self.client.consultarVariedadesClasesTabaco(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('variedad', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarRetencionesTabacaleras(self, sep="||"):
        "Retorna un listado de retenciones tabacaleras con código y descripción"
        ret = self.client.consultarRetencionesTabacaleras(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('retencion', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarDepositosAcopio(self, sep="||"):
        "Retorna los depósitos de acopio pertenencientes al contribuyente"
        ret = self.client.consultarDepositosAcopio(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('acopio', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarPuntosVentas(self, sep="||"):
        "Retorna los puntos de ventas autorizados para la utilizacion de WS"
        ret = self.client.consultarPuntosVentas(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuit': self.Cuit, },
                            )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('puntoVenta', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]


    def MostrarPDF(self, archivo, imprimir=False):
        try:
            if sys.platform=="linux2":
                os.system("evince ""%s""" % archivo)
            else:
                operation = imprimir and "print" or ""
                os.startfile(archivo, operation)
            return True
        except Exception as e:
            self.Excepcion = str(e)
            return False


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = WSLTV.InstallDir = get_install_dir()


if __name__ == '__main__':
    if '--ayuda' in sys.argv:
        print(LICENCIA)
        print(AYUDA)
        sys.exit(0)
    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato in []:
            comienzo = 1
            print("=== %s ===" % msg)
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                dec = len(fmt)>3 and fmt[3] or (tipo=='I' and '2' or '')
                print(" * Campo: %-20s Posición: %3d Longitud: %4d Tipo: %s Decimales: %s" % (
                    clave, comienzo, longitud, tipo, dec))
                comienzo += longitud
        sys.exit(0)

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSLTV)
        sys.exit(0)

    import csv
    from configparser import SafeConfigParser

    from .wsaa import WSAA

    try:

        if "--version" in sys.argv:
            print("Versión: ", __version__)

        if len(sys.argv)>1 and sys.argv[1].endswith(".ini"):
            CONFIG_FILE = sys.argv[1]
            print("Usando configuracion:", CONFIG_FILE)

        config = SafeConfigParser()
        config.read(CONFIG_FILE)
        CERT = config.get('WSAA','CERT')
        PRIVATEKEY = config.get('WSAA','PRIVATEKEY')
        CUIT = config.get('WSLTV','CUIT')
        ENTRADA = config.get('WSLTV','ENTRADA')
        SALIDA = config.get('WSLTV','SALIDA')

        if config.has_option('WSAA','URL') and not HOMO:
            WSAA_URL = config.get('WSAA','URL')
        else:
            WSAA_URL = None #wsaa.WSAAURL
        if config.has_option('WSLTV','URL') and not HOMO:
            WSLTV_URL = config.get('WSLTV','URL')
        else:
            WSLTV_URL = WSDL

        PROXY = config.has_option('WSAA', 'PROXY') and config.get('WSAA', 'PROXY') or None
        CACERT = config.has_option('WSAA', 'CACERT') and config.get('WSAA', 'CACERT') or None
        WRAPPER = config.has_option('WSAA', 'WRAPPER') and config.get('WSAA', 'WRAPPER') or None

        if config.has_section('DBF'):
            conf_dbf = dict(config.items('DBF'))
            if DEBUG: print("conf_dbf", conf_dbf)
        else:
            conf_dbf = {}

        DEBUG = '--debug' in sys.argv
        XML = '--xml' in sys.argv

        if DEBUG:
            print("Usando Configuración:")
            print("WSAA_URL:", WSAA_URL)
            print("WSLTV_URL:", WSLTV_URL)
            print("CACERT", CACERT)
            print("WRAPPER", WRAPPER)
        # obteniendo el TA
        from .wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wsltv", CERT, PRIVATEKEY, wsdl=WSAA_URL,
                               proxy=PROXY, wrapper=WRAPPER, cacert=CACERT)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)

        # cliente soap del web service
        wsltv = WSLTV()
        wsltv.LanzarExcepciones = True
        wsltv.Conectar(url=WSLTV_URL, proxy=PROXY, wrapper=WRAPPER, cacert=CACERT)
        wsltv.SetTicketAcceso(ta)
        wsltv.Cuit = CUIT

        if '--dummy' in sys.argv:
            ret = wsltv.Dummy()
            print("AppServerStatus", wsltv.AppServerStatus)
            print("DbServerStatus", wsltv.DbServerStatus)
            print("AuthServerStatus", wsltv.AuthServerStatus)
            ##sys.exit(0)

        if '--autorizar' in sys.argv:

            # genero una liquidación de ejemplo:

            tipo_cbte = 150
            pto_vta = 2002

            if not '--prueba' in sys.argv:
                # consulto el último número de orden emitido:
                ok = wsltv.ConsultarUltimoComprobante(tipo_cbte, pto_vta)
                if ok:
                    nro_cbte = wsltv.NroComprobante + 1
            else:
                nro_cbte = 1

            # datos de la cabecera:
            fecha = '2016-01-01'
            cod_deposito_acopio = 207
            tipo_compra = 'CPS'
            variedad_tabaco = 'BR'
            cod_provincia_origen_tabaco = 1
            puerta = 22
            nro_tarjeta = 6569866
            horas = 12
            control = "FFAA"
            nro_interno = "77888"

            # cargo la liquidación:
            wsltv.CrearLiquidacion(tipo_cbte, pto_vta, nro_cbte, fecha,
                cod_deposito_acopio, tipo_compra,
                variedad_tabaco, cod_provincia_origen_tabaco,
                puerta, nro_tarjeta, horas, control,
                nro_interno, iibb_emisor=None)

            wsltv.AgregarCondicionVenta(codigo=99, descripcion="otra")

            # datos del receptor:
            cuit = 20111111112
            iibb = 123456
            nro_socio = 11223
            nro_fet = 22
            wsltv.AgregarReceptor(cuit, iibb, nro_socio, nro_fet)

            # datos romaneo:
            nro_romaneo = 321
            fecha_romaneo = "2015-12-10"
            wsltv.AgregarRomaneo(nro_romaneo, fecha_romaneo)
            # fardo:
            cod_trazabilidad = 355
            clase_tabaco = 4
            peso= 900
            wsltv.AgregarFardo(cod_trazabilidad, clase_tabaco, peso)

            # precio clase:
            precio = 190
            wsltv.AgregarPrecioClase(clase_tabaco, precio)

            # retencion:
            descripcion = "otra retencion"
            cod_retencion = 99
            importe = 12
            wsltv.AgregarRetencion(cod_retencion, descripcion, importe)

            # tributo:
            codigo_tributo = 99
            descripcion = "Ganancias"
            base_imponible = 15000
            alicuota = 8
            importe = 1200
            wsltv.AgregarTributo(codigo_tributo, descripcion, base_imponible, alicuota, importe)

            if '--testing' in sys.argv:
                # mensaje de prueba (no realiza llamada remota),
                # usar solo si no está operativo, cargo respuesta:
                wsltv.LoadTestXML("tests/xml/wsltv_aut_test.xml")

            print("Liquidacion: pto_vta=%s nro_cbte=%s tipo_cbte=%s" % (
                    wsltv.solicitud['liquidacion']['puntoVenta'],
                    wsltv.solicitud['liquidacion']['nroComprobante'],
                    wsltv.solicitud['liquidacion']['tipoComprobante'],
                    ))

            if not '--dummy' in sys.argv:
                print("Autorizando...")
                ret = wsltv.AutorizarLiquidacion()

            if wsltv.Excepcion:
                print("EXCEPCION:", wsltv.Excepcion, file=sys.stderr)
                if DEBUG: print(wsltv.Traceback, file=sys.stderr)
            print("Errores:", wsltv.Errores)
            print("CAE", wsltv.CAE)
            print("FechaLiquidacion", wsltv.FechaLiquidacion)
            print("NroComprobante", wsltv.NroComprobante)
            print("ImporteNeto", wsltv.ImporteNeto)
            print("AlicuotaIVA", wsltv.AlicuotaIVA)
            print("ImporteIVA", wsltv.ImporteIVA)
            print("Subtotal", wsltv.Subtotal)
            print("TotalRetenciones", wsltv.TotalRetenciones)
            print("TotalTributos", wsltv.TotalTributos)
            print("Total", wsltv.Total)
            if '--testing' in sys.argv:
                assert wsltv.CAE == "85523002502850"
                assert wsltv.Total == 205685.46
                assert wsltv.GetParametro("fecha") == "2016-01-01"
                assert wsltv.GetParametro("peso_total_fardos_kg") == "900"
                assert wsltv.GetParametro("cantidad_total_fardos") == "1"
                assert wsltv.GetParametro("emisor", "domicilio") == 'Peru 100'
                assert wsltv.GetParametro("emisor", "razon_social") == 'JOCKER'
                assert wsltv.GetParametro("receptor", "domicilio") == 'Calle 1'
                assert wsltv.GetParametro("receptor", "razon_social") == 'CUIT PF de Prueba gen\xe9rica'
                assert wsltv.GetParametro("romaneos", 0, "detalle_clase", 0, "cantidad_fardos") == "1"
                assert wsltv.GetParametro("romaneos", 0, "detalle_clase", 0, "cod_clase") == "4"
                assert wsltv.GetParametro("romaneos", 0, "detalle_clase", 0, "importe") == "171000.0"
                assert wsltv.GetParametro("romaneos", 0, "detalle_clase", 0, "peso_fardos_kg") == "900"
                assert wsltv.GetParametro("romaneos", 0, "detalle_clase", 0, "precio_x_kg_fardo") == "190.0"
                assert wsltv.GetParametro("romaneos", 0, "nro_romaneo") == "321"
                assert wsltv.GetParametro("romaneos", 0, "fecha_romaneo") == "2015-12-10"

            if DEBUG:
                pprint.pprint(wsltv.params_out)


        if '--consultar' in sys.argv:
            tipo_cbte = 151
            pto_vta = 1
            nro_cbte = 0
            try:
                tipo_cbte = sys.argv[sys.argv.index("--consultar") + 1]
                pto_vta = sys.argv[sys.argv.index("--consultar") + 2]
                nro_cbte = sys.argv[sys.argv.index("--consultar") + 3]
            except IndexError:
                pass
            if '--testing' in sys.argv:
                # mensaje de prueba (no realiza llamada remota),
                # usar solo si no está operativo, cargo prueba:
                wsltv.LoadTestXML("tests/xml/wsltv_cons_test.xml")
            print("Consultando: tipo_cbte=%s pto_vta=%s nro_cbte=%s" % (tipo_cbte, pto_vta, nro_cbte))
            ret = wsltv.ConsultarLiquidacion(tipo_cbte, pto_vta, nro_cbte)
            print("CAE", wsltv.CAE)
            print("Errores:", wsltv.Errores)

            if DEBUG:
                pprint.pprint(wsltv.params_out)

            if '--mostrar' in sys.argv and pdf:
                wsltv.MostrarPDF(archivo=pdf,
                                 imprimir='--imprimir' in sys.argv)

        if '--ult' in sys.argv:
            tipo_cbte = 151
            pto_vta = 1
            try:
                tipo_cbte = sys.argv[sys.argv.index("--ult") + 1]
                pto_vta = sys.argv[sys.argv.index("--ult") + 2]
            except IndexError:
                pass

            print("Consultando ultimo nro_cbte para pto_vta=%s" % pto_vta, end=' ')
            ret = wsltv.ConsultarUltimoComprobante(tipo_cbte, pto_vta)
            if wsltv.Excepcion:
                print("EXCEPCION:", wsltv.Excepcion, file=sys.stderr)
                if DEBUG: print(wsltv.Traceback, file=sys.stderr)
            print("Ultimo Nro de Comprobante", wsltv.NroComprobante)
            print("Errores:", wsltv.Errores)
            sys.exit(0)

        # Recuperar parámetros:

        if '--provincias' in sys.argv:
            ret = wsltv.ConsultarProvincias()
            print("\n".join(ret))

        if '--condicionesventa' in sys.argv:
            ret = wsltv.ConsultarCondicionesVenta()
            print("\n".join(ret))

        if '--tributos' in sys.argv:
            ret = wsltv.ConsultarTributos()
            print("\n".join(ret))

        if '--retenciones' in sys.argv:
            ret = wsltv.ConsultarRetencionesTabacaleras()
            print("\n".join(ret))

        if '--variedades' in sys.argv:
            ret = wsltv.ConsultarVariedadesClasesTabaco()
            print("\n".join(ret))

        if '--depositos' in sys.argv:
            ret = wsltv.ConsultarDepositosAcopio()
            print("\n".join(ret))

        if '--puntosventa' in sys.argv:
            ret = wsltv.ConsultarPuntosVentas()
            print("\n".join(ret))

        print("hecho.")

    except SoapFault as e:
        print("Falla SOAP:", e.faultcode, e.faultstring.encode("ascii","ignore"), file=sys.stderr)
        sys.exit(3)
    except Exception as e:
        try:
            print(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[0], file=sys.stderr)
        except:
            print("Excepción no disponible: ", type(e), file=sys.stderr)
        if DEBUG:
            raise
        sys.exit(5)
    finally:
        if XML:
            open("wsltv_request.xml", "w").write(wsltv.client.xml_request)
            open("wsltv_response.xml", "w").write(wsltv.client.xml_response)



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo para Trazabilidad de Medicamentos ANMAT - PAMI - INSSJP Disp. 3683/11
seg�n Especificaci�n T�cnica para Pruebas de Servicios v2 (2013)"""

# Informaci�n adicional y documentaci�n:
# http://www.sistemasagiles.com.ar/trac/wiki/TrazabilidadMedicamentos

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2011 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.16b"

import os
import socket
import sys
import datetime
import time
import traceback
import py3simplesoap.client
from py3simplesoap.client import SoapClient, SoapFault, parse_proxy, \
                                set_http_wrapper
from py3simplesoap.simplexml import SimpleXMLElement
from cStringIO import StringIO

# importo funciones compartidas:
from utils import (
    leer, escribir, leer_dbf, guardar_dbf, N, A, I, json,
    dar_nombre_campo_dbf, get_install_dir, BaseWS,
    inicializar_y_capturar_excepciones
)

HOMO = False
TYPELIB = False

WSDL = "https://servicios.pami.org.ar/trazamed.WebService?wsdl"
LOCATION = "https://servicios.pami.org.ar/trazamed.WebService"
# WSDL = "https://trazabilidad.pami.org.ar:9050/trazamed.WebService?wsdl"

# Formato de MedicamentosDTO, MedicamentosDTODHSerie, MedicamentosDTOFraccion
MEDICAMENTOS = [
    ('f_evento', 10, A),                # formato DD/MM/AAAA
    ('h_evento', 5, A),                 # formato HH:MM
    ('gln_origen', 13, A),
    ('gln_destino', 13, A),
    ('n_remito', 20, A),
    ('n_factura', 20, A),
    ('vencimiento', 10, A),
    ('gtin', 14, A),
    ('lote', 20, A),
    ('numero_serial', 20, A),
    ('desde_numero_serial', 20, A),     # sendMedicamentosDHSerie
    ('hasta_numero_serial', 20, A),     # sendMedicamentosDHSerie
    ('id_obra_social', 9, N),
    ('id_evento', 3, N),
    ('cuit_origen', 11, A),
    ('cuit_destino', 11, A),
    ('apellido', 50, A),
    ('nombres', 100, A),
    ('tipo_documento', 2, N),           # 96: DNI,80: CUIT
    ('n_documento', 10, A),
    ('sexo', 1, A),                     # M o F
    ('direccion', 100, A),
    ('numero', 10, A),
    ('piso', 5, A),
    ('depto', 5, A),
    ('localidad', 50, A),
    ('provincia', 100, A),
    ('n_postal', 8, A),
    ('fecha_nacimiento', 100, A),
    ('telefono', 30, A),
    ('nro_asociado', 30, A),
    ('cantidad', 3, N),                 # sendMedicamentosFraccion
    ('codigo_transaccion', 14, A),
]

# Formato para TransaccionPlainWS (getTransaccionesNoConfirmadas)
TRANSACCIONES = [
    ('_id_transaccion', 14, A),
    ('_id_transaccion_global', 14, A),
    ('_f_evento', 10, A),
    ('_f_transaccion', 16, A),          # formato DD/MM/AAAA HH:MM
    ('_gtin', 14, A),
    ('_lote', 20, A),
    ('_numero_serial', 20, A),
    ('_nombre', 200, A),
    ('_d_evento', 100, A),
    ('_gln_origen', 13, A),
    ('_razon_social_origen', 200, A),
    ('_gln_destino', 13, A),
    ('_razon_social_destino', 200, A),
    ('_n_remito', 20, A),
    ('_n_factura', 20, A),
    ('_vencimiento', 10, A),
    ('_id_evento', 3, N),               # agregado el 30/01/2014
]

# Formato para Errores
ERRORES = [
    ('c_error', 4, A),                 # c�digo
    ('d_error', 250, A),               # descripci�n
    ]


class TrazaMed(BaseWS):
    "Interfaz para el WebService de Trazabilidad de Medicamentos ANMAT - \
    PAMI - INSSJP"
    _public_methods_ = ['SendMedicamentos',
                        'SendCancelacTransacc', 'SendCancelacTransaccParcial',
                        'SendMedicamentosDHSerie',
                        'SendMedicamentosFraccion',
                        'SendConfirmaTransacc', 'SendAlertaTransacc',
                        'GetTransaccionesNoConfirmadas',
                        'GetEnviosPropiosAlertados', 'GetConsultaStock',
                        'GetTransaccionesWS', 'GetCatalogoElectronicoByGTIN',
                        'Conectar', 'LeerError', 'LeerTransaccion',
                        'SetUsername', 'SetPassword',
                        'SetParametro', 'GetParametro',
                        'GetCodigoTransaccion', 'GetResultado', 'LoadTestXML']

    _public_attrs_ = [
        'Username', 'Password',
        'CodigoTransaccion', 'Errores', 'Resultado',
        'XmlRequest', 'XmlResponse',
        'Version', 'InstallDir',
        'Traceback', 'Excepcion', 'LanzarExcepciones',
        'CantPaginas', 'HayError', 'TransaccionPlainWS',
        ]

    _reg_progid_ = "TrazaMed"
    _reg_clsid_ = "{8472867A-AE6F-487F-8554-C2C896CFFC3E}"

    if TYPELIB:
        _typelib_guid_ = '{F992EB7E-AFBD-41BB-B717-5693D3A2BADB}'
        _typelib_version_ = 1, 4
        _com_interfaces_ = ['ITrazaMed']

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s %s" % (__version__, HOMO and 'Homologaci�n' or '',
                            py3simplesoap.client.__version__)

    def __init__(self, reintentos=1):
        self.Username = self.Password = None
        self.TransaccionPlainWS = []
        BaseWS.__init__(self, reintentos)

    def inicializar(self):
        BaseWS.inicializar(self)
        self.CodigoTransaccion = self.Errores = self.Resultado = None
        self.Resultado = ''
        self.Errores = []   # lista de strings para la interfaz
        self.errores = []   # lista de diccionarios (uso interno)
        self.CantPaginas = self.HayError = None

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        self.errores = ret.get('errores', [])
        self.Errores = ["%s: %s" % (it['_c_error'], it['_d_error'])
                        for it in ret.get('errores', [])]
        self.Resultado = ret.get('resultado')

    def Conectar(
        self, cache=None, wsdl=None, proxy="", wrapper=None,
        cacert=None, timeout=30
    ):
        # Conecto usando el m�todo estandard:
        ok = BaseWS.Conectar(self, cache, wsdl, proxy, wrapper, cacert,
                             timeout, soap_server="jetty")

        if ok:
            # si el archivo es local, asumo que ya esta corregido:
            if not self.wsdl.startswith("file"):
                # corrijo ubicaci�n del servidor (localhost:9050 en el WSDL)
                location = self.wsdl[:-5]
                if 'IWebServiceService' in self.client.services:
                    # version 1
                    ws = self.client.services['IWebServiceService']
                else:
                    ws = self.client.services['IWebService']   # version 2
                ws['ports']['IWebServicePort']['location'] = location

            # Establecer credenciales de seguridad:
            self.client['wsse:Security'] = {
                'wsse:UsernameToken': {
                    'wsse:Username': self.Username,
                    'wsse:Password': self.Password,
                    }
                }
        return ok

    @inicializar_y_capturar_excepciones
    def SendMedicamentos(self, usuario, password,
                         f_evento, h_evento, gln_origen, gln_destino,
                         n_remito, n_factura, vencimiento, gtin, lote,
                         numero_serial, id_obra_social, id_evento,
                         cuit_origen='', cuit_destino='', apellido='',
                         nombres='', tipo_documento='', n_documento='',
                         sexo='', direccion='', numero='', piso='',
                         depto='', localidad='', provincia='',
                         n_postal='', fecha_nacimiento='', telefono='',
                         nro_asociado=None,
                         ):
        "Realiza el registro de una transacci�n de medicamentos. "
        # creo los par�metros para esta llamada
        params = {
            'f_evento': f_evento,
            'h_evento': h_evento,
            'gln_origen': gln_origen,
            'gln_destino': gln_destino,
            'n_remito': n_remito,
            'n_factura': n_factura,
            'vencimiento': vencimiento,
            'gtin': gtin,
            'lote': lote,
            'numero_serial': numero_serial,
            'id_obra_social': id_obra_social or None,
            'id_evento': id_evento,
            'cuit_origen': cuit_origen,
            'cuit_destino': cuit_destino,
            'apellido': apellido,
            'nombres': nombres,
            'tipo_documento': tipo_documento,
            'n_documento': n_documento,
            'sexo': sexo,
            'direccion': direccion,
            'numero': numero,
            'piso': piso,
            'depto': depto,
            'localidad': localidad,
            'provincia': provincia,
            'n_postal': n_postal,
            'fecha_nacimiento': fecha_nacimiento,
            'telefono': telefono,
            'nro_asociado': nro_asociado,
        }
        res = self.client.sendMedicamentos(
            arg0=params,
            arg1=usuario,
            arg2=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret['codigoTransaccion']
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendMedicamentosFraccion(
        self, usuario, password,
        f_evento, h_evento, gln_origen, gln_destino,
        n_remito, n_factura, vencimiento, gtin, lote,
        numero_serial, id_obra_social, id_evento,
        cuit_origen='', cuit_destino='', apellido='', nombres='',
        tipo_documento='', n_documento='', sexo='',
        direccion='', numero='', piso='', depto='', localidad='', provincia='',
        n_postal='', fecha_nacimiento='', telefono='',
        nro_asociado=None, cantidad=None,
    ):
        "Realiza el registro de una transacci�n de medicamentos fraccionados"
        # creo los par�metros para esta llamada
        params = {
            'f_evento': f_evento,
            'h_evento': h_evento,
            'gln_origen': gln_origen,
            'gln_destino': gln_destino,
            'n_remito': n_remito,
            'n_factura': n_factura,
            'vencimiento': vencimiento,
            'gtin': gtin,
            'lote': lote,
            'numero_serial': numero_serial,
            'id_obra_social': id_obra_social or None,
            'id_evento': id_evento,
            'cuit_origen': cuit_origen,
            'cuit_destino': cuit_destino,
            'apellido': apellido,
            'nombres': nombres,
            'tipo_documento': tipo_documento,
            'n_documento': n_documento,
            'sexo': sexo,
            'direccion': direccion,
            'numero': numero,
            'piso': piso,
            'depto': depto,
            'localidad': localidad,
            'provincia': provincia,
            'n_postal': n_postal,
            'fecha_nacimiento': fecha_nacimiento,
            'telefono': telefono,
            'nro_asociado': nro_asociado,
            'cantidad': cantidad,
        }
        res = self.client.sendMedicamentosFraccion(
            arg0=params,
            arg1=usuario,
            arg2=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret['codigoTransaccion']
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendMedicamentosDHSerie(
        self, usuario, password,
        f_evento, h_evento, gln_origen, gln_destino,
        n_remito, n_factura, vencimiento, gtin, lote,
        desde_numero_serial, hasta_numero_serial,
        id_obra_social, id_evento,
        cuit_origen='', cuit_destino='', apellido='', nombres='',
        tipo_documento='', n_documento='', sexo='',
        direccion='', numero='', piso='', depto='', localidad='', provincia='',
        n_postal='', fecha_nacimiento='', telefono='',
        nro_asociado=None,
    ):
        "Env�a un lote de medicamentos informando el desde-hasta n�mero de \
            serie"
        # creo los par�metros para esta llamada
        params = {
            'f_evento': f_evento,
            'h_evento': h_evento,
            'gln_origen': gln_origen,
            'gln_destino': gln_destino,
            'n_remito': n_remito,
            'n_factura': n_factura,
            'vencimiento': vencimiento,
            'gtin': gtin,
            'lote': lote,
            'desde_numero_serial': desde_numero_serial,
            'hasta_numero_serial': hasta_numero_serial,
            'id_obra_social': id_obra_social or None,
            'id_evento': id_evento,
            'cuit_origen': cuit_origen,
            'cuit_destino': cuit_destino,
            'apellido': apellido,
            'nombres': nombres,
            'tipo_documento': tipo_documento,
            'n_documento': n_documento,
            'sexo': sexo,
            'direccion': direccion,
            'numero': numero,
            'piso': piso,
            'depto': depto,
            'localidad': localidad,
            'provincia': provincia,
            'n_postal': n_postal,
            'fecha_nacimiento': fecha_nacimiento,
            'telefono': telefono,
            'nro_asociado': nro_asociado,
        }
        res = self.client.sendMedicamentosDHSerie(
            arg0=params,
            arg1=usuario,
            arg2=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret['codigoTransaccion']
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendCancelacTransacc(self, usuario, password, codigo_transaccion):
        " Realiza la cancelaci�n de una transacci�n"
        res = self.client.sendCancelacTransacc(
            arg0=codigo_transaccion,
            arg1=usuario,
            arg2=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendCancelacTransaccParcial(
        self, usuario, password, codigo_transaccion,
        gtin_medicamento=None, numero_serial=None
    ):
        " Realiza la cancelaci�n parcial de una transacci�n"
        res = self.client.sendCancelacTransaccParcial(
            arg0=codigo_transaccion,
            arg1=usuario,
            arg2=password,
            arg3=gtin_medicamento,
            arg4=numero_serial,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendConfirmaTransacc(
        self, usuario, password, p_ids_transac, f_operacion
    ):
        "Confirma la recepci�n de un medicamento"
        res = self.client.sendConfirmaTransacc(
            arg0=usuario,
            arg1=password,
            arg2={'p_ids_transac': p_ids_transac, 'f_operacion': f_operacion},
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('id_transac_asociada')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendAlertaTransacc(self, usuario, password, p_ids_transac_ws):
        "Alerta un medicamento, acci�n contraria a �confirmar la \
            transacci�n�."
        res = self.client.sendAlertaTransacc(
            arg0=usuario,
            arg1=password,
            arg2=p_ids_transac_ws,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('id_transac_asociada')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def GetTransaccionesNoConfirmadas(
        self, usuario, password,
        p_id_transaccion_global=None, id_agente_informador=None,
        id_agente_origen=None, id_agente_destino=None,
        id_medicamento=None, id_evento=None,
        fecha_desde_op=None, fecha_hasta_op=None,
        fecha_desde_t=None, fecha_hasta_t=None,
        fecha_desde_v=None, fecha_hasta_v=None,
        n_remito=None, n_factura=None,
        estado=None, lote=None, numero_serial=None,
    ):
        "Trae un listado de las transacciones que no est�n confirmadas"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if p_id_transaccion_global is not None:
            kwargs['arg2'] = p_id_transaccion_global
        if id_agente_informador is not None:
            kwargs['arg3'] = id_agente_informador
        if id_agente_origen is not None:
            kwargs['arg4'] = id_agente_origen
        if id_agente_destino is not None:
            kwargs['arg5'] = id_agente_destino
        if id_medicamento is not None:
            kwargs['arg6'] = id_medicamento
        if id_evento is not None:
            kwargs['arg7'] = id_evento
        if fecha_desde_op is not None:
            kwargs['arg8'] = fecha_desde_op
        if fecha_hasta_op is not None:
            kwargs['arg9'] = fecha_hasta_op
        if fecha_desde_t is not None:
            kwargs['arg10'] = fecha_desde_t
        if fecha_hasta_t is not None:
            kwargs['arg11'] = fecha_hasta_t
        if fecha_desde_v is not None:
            kwargs['arg12'] = fecha_desde_v
        if fecha_hasta_v is not None:
            kwargs['arg13'] = fecha_hasta_v
        if n_remito is not None:
            kwargs['arg14'] = n_remito
        if n_factura is not None:
            kwargs['arg15'] = n_factura
        if estado is not None:
            kwargs['arg16'] = estado
        if lote is not None:
            kwargs['arg17'] = lote
        if numero_serial is not None:
            kwargs['arg18'] = numero_serial

        # llamo al webservice
        res = self.client.getTransaccionesNoConfirmadas(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.TransaccionPlainWS = [it for it in ret.get('list', [])]
        return True

    def LeerTransaccion(self):
        "Recorro TransaccionPlainWS devuelto por GetTransaccionesNoConfirmadas"
        # usar GetParametro para consultar el valor retornado por el
        # webservice

        if self.TransaccionPlainWS:
            # extraigo el primer item
            self.params_out = self.TransaccionPlainWS.pop(0)
            return True
        else:
            # limpio los par�metros
            self.params_out = {}
            return False

    def LeerError(self):
        "Recorro los errores devueltos y devuelvo el primero si existe"

        if self.Errores:
            # extraigo el primer item
            er = self.Errores.pop(0)
            return er
        else:
            return ""

    @inicializar_y_capturar_excepciones
    def GetEnviosPropiosAlertados(
        self, usuario, password,
        p_id_transaccion_global=None, id_agente_informador=None,
        id_agente_origen=None, id_agente_destino=None,
        id_medicamento=None, id_evento=None,
        fecha_desde_op=None, fecha_hasta_op=None,
        fecha_desde_t=None, fecha_hasta_t=None,
        fecha_desde_v=None, fecha_hasta_v=None,
        n_remito=None, n_factura=None,
    ):
        "Obtiene las distribuciones y env�os propios que han sido alertados"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if p_id_transaccion_global is not None:
            kwargs['arg2'] = p_id_transaccion_global
        if id_agente_informador is not None:
            kwargs['arg3'] = id_agente_informador
        if id_agente_origen is not None:
            kwargs['arg4'] = id_agente_origen
        if id_agente_destino is not None:
            kwargs['arg5'] = id_agente_destino
        if id_medicamento is not None:
            kwargs['arg6'] = id_medicamento
        if id_evento is not None:
            kwargs['arg7'] = id_evento
        if fecha_desde_op is not None:
            kwargs['arg8'] = fecha_desde_op
        if fecha_hasta_op is not None:
            kwargs['arg9'] = fecha_hasta_op
        if fecha_desde_t is not None:
            kwargs['arg10'] = fecha_desde_t
        if fecha_hasta_t is not None:
            kwargs['arg11'] = fecha_hasta_t
        if fecha_desde_v is not None:
            kwargs['arg12'] = fecha_desde_v
        if fecha_hasta_v is not None:
            kwargs['arg13'] = fecha_hasta_v
        if n_remito is not None:
            kwargs['arg14'] = n_remito
        if n_factura is not None:
            kwargs['arg15'] = n_factura

        # llamo al webservice
        res = self.client.getEnviosPropiosAlertados(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.TransaccionPlainWS = [it for it in ret.get('list', [])]
        return True

    @inicializar_y_capturar_excepciones
    def GetTransaccionesWS(
        self, usuario, password,
        p_id_transaccion_global=None,
        id_agente_origen=None, id_agente_destino=None,
        id_medicamento=None, id_evento=None,
        fecha_desde_op=None, fecha_hasta_op=None,
        fecha_desde_t=None, fecha_hasta_t=None,
        fecha_desde_v=None, fecha_hasta_v=None,
        n_remito=None, n_factura=None,
        id_estado=None, nro_pag=None,
    ):
        "Obtiene los movimientos realizados y permite filtros de b�squeda"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if p_id_transaccion_global is not None:
            kwargs['arg2'] = p_id_transaccion_global
        if id_agente_origen is not None:
            kwargs['arg3'] = id_agente_origen
        if id_agente_destino is not None:
            kwargs['arg4'] = id_agente_destino
        if id_medicamento is not None:
            kwargs['arg5'] = id_medicamento
        if id_evento is not None:
            kwargs['arg6'] = id_evento
        if fecha_desde_op is not None:
            kwargs['arg7'] = fecha_desde_op
        if fecha_hasta_op is not None:
            kwargs['arg8'] = fecha_hasta_op
        if fecha_desde_t is not None:
            kwargs['arg9'] = fecha_desde_t
        if fecha_hasta_t is not None:
            kwargs['arg10'] = fecha_hasta_t
        if fecha_desde_v is not None:
            kwargs['arg11'] = fecha_desde_v
        if fecha_hasta_v is not None:
            kwargs['arg12'] = fecha_hasta_v
        if n_remito is not None:
            kwargs['arg13'] = n_remito
        if n_factura is not None:
            kwargs['arg14'] = n_factura
        if id_estado is not None:
            kwargs['arg15'] = id_estado
        if nro_pag is not None:
            kwargs['arg16'] = nro_pag

        # llamo al webservice
        res = self.client.getTransaccionesWS(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.TransaccionPlainWS = [it for it in ret.get('list', [])]
        return True

    @inicializar_y_capturar_excepciones
    def GetCatalogoElectronicoByGTIN(
        self, usuario, password,
        cuit_fabricante=None, gtin=None, descripcion=None,
        id_monodroga=None,
    ):
        "Obtiene el Cat�logo Electr�nico de Medicamentos"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if cuit_fabricante is not None:
            kwargs['arg2'] = cuit_fabricante
        if gtin is not None:
            kwargs['arg3'] = gtin
        if descripcion is not None:
            kwargs['arg4'] = descripcion
        if id_monodroga is not None:
            kwargs['arg5'] = id_monodroga

        # llamo al webservice
        res = self.client.getCatalogoElectronicoByGTIN(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.params_out = dict(
                [(i, it) for i, it in enumerate(ret.get('list', []))])
            return len(self.params_out)
        else:
            return 0

    @inicializar_y_capturar_excepciones
    def GetConsultaStock(
        self, usuario, password,
        id_medicamento=None, id_agente=None, descripcion=None,
        cantidad=None, presentacion=None,
        lote=None, numero_serial=None,
        nro_pag=1, cant_reg=100,
    ):
        "Permite consultar el stock actual del agente."

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if id_medicamento is not None:
            kwargs['arg2'] = id_medicamento
        if id_agente is not None:
            kwargs['arg3'] = id_agente
        if descripcion is not None:
            kwargs['arg4'] = descripcion
        if cantidad is not None:
            kwargs['arg5'] = cantidad
        if presentacion is not None:
            kwargs['arg6'] = presentacion
        if lote is not None:
            kwargs['arg7'] = lote
        if numero_serial is not None:
            kwargs['arg8'] = numero_serial
        if nro_pag is not None:
            kwargs['arg9'] = nro_pag
        if cant_reg is not None:
            kwargs['arg10'] = cant_reg

        # llamo al webservice
        res = self.client.getConsultaStock(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.params_out = dict(
                [(i, it) for i, it in enumerate(ret.get('list', []))])
            return len(self.params_out)
        else:
            return 0

    def SetUsername(self, username):
        "Establezco el nombre de usuario"
        self.Username = username

    def SetPassword(self, password):
        "Establezco la contrase�a"
        self.Password = password

    def GetCodigoTransaccion(self):
        "Devuelvo el c�digo de transacci�n"
        return self.CodigoTransaccion

    def GetResultado(self):
        "Devuelvo el resultado"
        return self.Resultado


def main():
    "Funci�n principal de pruebas (obtener CAE)"
    import os
    import time
    import sys
    global WSDL, LOCATION

    DEBUG = '--debug' in sys.argv
    print(DEBUG)

    ws = TrazaMed()

    ws.Username = 'testwservice'
    ws.Password = 'testwservicepsw'

    if '--prod' in sys.argv and not HOMO:
        WSDL = "https://trazabilidad.pami.org.ar:9050/trazamed.WebService"
        print("Usando WSDL:", WSDL)
        sys.argv.pop(sys.argv.index("--prod"))

    # Inicializo las variables y estructuras para el archivo de intercambio:
    medicamentos = []
    transacciones = []
    errores = []
    formatos = [
        ('Medicamentos', MEDICAMENTOS, medicamentos),
        ('Transacciones', TRANSACCIONES, transacciones),
        ('Errores', ERRORES, errores),
    ]

    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato, lista in formatos:
            comienzo = 1
            print("=== %s ===" % msg)
            print("|| %-25s || %-12s || %-5s || %-4s || %-10s ||" % (
                "Nombre", "Tipo", "Long.", "Pos(txt)", "Campo(dbf)"))
            claves = []
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                clave_dbf = dar_nombre_campo_dbf(clave, claves)
                claves.append(clave_dbf)
                print("|| %-25s || %-12s || %5d ||   %4d   || %-10s ||" % (
                    clave, tipo, longitud, comienzo, clave_dbf))
                comienzo += longitud
        sys.exit(0)

    if '--cargar' in sys.argv:
        if '--dbf' in sys.argv:
            leer_dbf(formatos[:1], {})
        elif '--json' in sys.argv:
            for formato in formatos[:1]:
                archivo = open(formato[0].lower() + ".json", "r")
                d = json.load(archivo)
                formato[2].extend(d)
                archivo.close()
        else:
            for formato in formatos[:1]:
                archivo = open(formato[0].lower() + ".txt", "r")
                for linea in archivo:
                    d = leer(linea, formato[1])
                    formato[2].append(d)
                archivo.close()

    ws.Conectar("", WSDL)

    if ws.Excepcion:
        print(ws.Excepcion)
        print(ws.Traceback)
        sys.exit(-1)

    # Datos de pruebas:

    if '--test' in sys.argv:
        medicamentos.append(dict(
            f_evento=datetime.datetime.now().strftime("%d/%m/%Y"),
            h_evento=datetime.datetime.now().strftime("%H:%M"),
            gln_origen="9999999999918", gln_destino="glnws",
            n_remito="R000100001234", n_factura="A000100001234",
            vencimiento=(
                datetime.datetime.now()+datetime.timedelta(30)
            ).strftime("%d/%m/%Y"),
            gtin="GTIN1", lote=datetime.datetime.now().strftime("%Y"),
            numero_serial=int(time.time()*10),
            id_obra_social=None, id_evento=134,
            cuit_origen="20267565393", cuit_destino="20267565393",
            apellido="Reingart", nombres="Mariano",
            tipo_documento="96", n_documento="26756539", sexo="M",
            direccion="Saraza", numero="1234", piso="", depto="",
            localidad="Hurlingham", provincia="Buenos Aires",
            n_postal="1688", fecha_nacimiento="01/01/2000",
            telefono="5555-5555",
            nro_asociado="9999999999999",
            cantidad=None,
            desde_numero_serial=None, hasta_numero_serial=None,
            codigo_transaccion=None,
        ))
    if '--testfraccion' in sys.argv:
        medicamentos.append(dict(
            f_evento=datetime.datetime.now().strftime("%d/%m/%Y"),
            h_evento=datetime.datetime.now().strftime("%H:%M"),
            gln_origen="9999999999918", gln_destino="glnws",
            n_remito="1234", n_factura="1234",
            vencimiento=(
                datetime.datetime.now()+datetime.timedelta(30)
            ).strftime("%d/%m/%Y"),
            gtin="GTIN1", lote=datetime.datetime.now().strftime("%Y"),
            numero_serial=int(time.time()*10),
            id_obra_social=None, id_evento=134,
            cuit_origen="20267565393", cuit_destino="20267565393",
            apellido="Reingart", nombres="Mariano",
            tipo_documento="96", n_documento="26756539", sexo="M",
            direccion="Saraza", numero="1234", piso="", depto="",
            localidad="Hurlingham", provincia="Buenos Aires",
            n_postal="1688", fecha_nacimiento="01/01/2000",
            telefono="5555-5555",
            nro_asociado="9999999999999",
            cantidad=5,
            desde_numero_serial=None, hasta_numero_serial=None,
            codigo_transaccion=None,
        ))
    if '--testdh' in sys.argv:
        medicamentos.append(dict(
            f_evento=datetime.datetime.now().strftime("%d/%m/%Y"),
            h_evento=datetime.datetime.now().strftime("%H:%M"),
            gln_origen="9999999999918", gln_destino="glnws",
            n_remito="1234", n_factura="1234",
            vencimiento=(
                datetime.datetime.now()+datetime.timedelta(30)
            ).strftime("%d/%m/%Y"),
            gtin="GTIN1", lote=datetime.datetime.now().strftime("%Y"),
            desde_numero_serial=int(time.time()*10)-1,
            hasta_numero_serial=int(time.time()*10)+1,
            id_obra_social=None, id_evento=134,
            nro_asociado="1234",
            cantidad=None, numero_serial=None,
            codigo_transaccion=None,
        ))

    # Opciones principales:

    if '--cancela' in sys.argv:
        if '--loadxml' in sys.argv:
            ws.LoadTestXML("trazamed_cancela_err.xml")  # cargo respuesta
        ws.SendCancelacTransacc(*sys.argv[sys.argv.index("--cancela")+1:])
    elif '--cancela_parcial' in sys.argv:
        ws.SendCancelacTransaccParcial(
            *sys.argv[sys.argv.index("--cancela_parcial")+1:])
    elif '--confirma' in sys.argv:
        if '--loadxml' in sys.argv:
            ws.LoadTestXML("trazamed_confirma.xml")  # cargo respuesta
            ok = ws.SendConfirmaTransacc(
                usuario="pruebasws", password="pruebasws",
                p_ids_transac="1", f_operacion="31-12-2013")
            if not ok:
                raise RuntimeError(ws.Excepcion)
        ws.SendConfirmaTransacc(*sys.argv[sys.argv.index("--confirma")+1:])
    elif '--alerta' in sys.argv:
        ws.SendAlertaTransacc(*sys.argv[sys.argv.index("--alerta")+1:])
    elif '--consulta' in sys.argv:
        if '--alertados' in sys.argv:
            ws.GetEnviosPropiosAlertados(
                *sys.argv[sys.argv.index("--alertados")+1:]
            )
        elif '--movimientos' in sys.argv:
            ws.GetTransaccionesWS(
                *sys.argv[sys.argv.index("--movimientos")+1:]
            )
        else:
            ws.GetTransaccionesNoConfirmadas(
                *sys.argv[sys.argv.index("--consulta")+1:]
                # usuario="pruebasws", password="pruebasws",
                # p_id_transaccion_global="1234",
                # id_agente_informador="1",
                # id_agente_origen="1",
                # id_agente_destino="1",
                # id_medicamento="1",
                # id_evento="1",
                # fecha_desde_op="01/01/2015",
                # fecha_hasta_op="31/12/2013",
                # fecha_desde_t="01/01/2013",
                # fecha_hasta_t="31/12/2013",
                # fecha_desde_v="01/04/2013",
                # fecha_hasta_v="30/04/2013",
                # n_factura=5, n_remito=6,
                # estado=1,
                # lote=88745,
                # numero_serial=894124788,
            )
        print("CantPaginas", ws.CantPaginas)
        print("HayError", ws.HayError)
        # print("TransaccionPlainWS", ws.TransaccionPlainWS)
        # parametros comunes de salida (columnas de la tabla):
        claves = [k for k, v, l in TRANSACCIONES]
        # extiendo la lista de resultado para el archivo de intercambio:
        transacciones.extend(ws.TransaccionPlainWS)
        # encabezado de la tabla:
        print("||", "||".join(["%s" % clave for clave in claves]), "||")
        # recorro los datos devueltos (TransaccionPlainWS):
        while ws.LeerTransaccion():
            for clave in claves:
                print("||", ws.GetParametro(clave))         # imprimo cada fila
            print("||")
    elif '--catalogo' in sys.argv:
        ret = ws.GetCatalogoElectronicoByGTIN(
                                *sys.argv[sys.argv.index("--catalogo")+1:]
                                )
        for catalogo in ws.params_out.values():
            print(catalogo)        # imprimo cada fila)
    elif '--stock' in sys.argv:
        ret = ws.GetConsultaStock(
            *sys.argv[sys.argv.index("--stock")+1:]
        )
        print(ret)
        print("\n".join([str(s) for s in ws.params_out.values()]))
    else:
        argv = [argv for argv in sys.argv if not argv.startswith("--")]
        if not medicamentos:
            if len(argv) > 16:
                if '--dh' in sys.argv:
                    ws.SendMedicamentosDHSerie(*argv[1:])
                elif '--fraccion' in sys.argv:
                    ws.SendMedicamentosFraccion(*argv[1:])
                else:
                    ws.SendMedicamentos(*argv[1:])
            else:
                print("ERROR: no se indicaron todos los par�metros requeridos")
        elif medicamentos:
            try:
                usuario, password = argv[1:3]
            except Exception as e:
                print(
                    "ADVERTENCIA: no se indico par�metros usuario y passoword"
                )
                print(e)
                usuario = password = "pruebasws"
            for i, med in enumerate(medicamentos):
                print("Procesando registro", i)
                del med['codigo_transaccion']
                if med.get("cantidad"):
                    del med["desde_numero_serial"]
                    del med["hasta_numero_serial"]
                    ws.SendMedicamentosFraccion(usuario, password, **med)
                elif med.get("desde_numero_serial"):
                    del med["cantidad"]
                    del med["numero_serial"]
                    ws.SendMedicamentosDHSerie(usuario, password, **med)
                else:
                    del med["cantidad"]
                    del med["desde_numero_serial"]
                    del med["hasta_numero_serial"]
                    ws.SendMedicamentos(usuario, password, **med)
                med['codigo_transaccion'] = ws.CodigoTransaccion
                errores.extend(ws.errores)
                print("|Resultado %5s|CodigoTransaccion %10s|Errores|%s|" % (
                    ws.Resultado,
                    ws.CodigoTransaccion,
                    '|'.join(ws.Errores or []),
                ))
        else:
            print("ERROR: no se especificaron medicamentos a informar")

    if not medicamentos:
        print("|Resultado %5s|CodigoTransaccion %10s|Errores|%s|" % (
            ws.Resultado,
            ws.CodigoTransaccion,
            '|'.join(ws.Errores or []),
        ))

    if ws.Excepcion:
        print(ws.Traceback)

    if '--grabar' in sys.argv:
        if '--dbf' in sys.argv:
            guardar_dbf(formatos, True, {})
        elif '--json' in sys.argv:
            for formato in formatos:
                archivo = open(formato[0].lower() + ".json", "w")
                json.dump(formato[2], archivo, sort_keys=True, indent=4)
                archivo.close()
        else:
            for formato in formatos:
                archivo = open(formato[0].lower() + ".txt", "w")
                for it in formato[2]:
                    archivo.write(escribir(it, formato[1]))
            archivo.close()


# busco el directorio de instalaci�n (global para que no cambie si
# usan otra dll)
INSTALL_DIR = TrazaMed.InstallDir = get_install_dir()


if __name__ == '__main__':

    # ajusto el encoding por defecto (si se redirije la salida)
    if not hasattr(sys.stdout, "encoding") or sys.stdout.encoding is None:
        import codecs
        import locale
        sys.stdout = codecs.getwriter(
            locale.getpreferredencoding())(sys.stdout, "replace")
        sys.stderr = codecs.getwriter(
            locale.getpreferredencoding())(sys.stderr, "replace")

    if '--register' in sys.argv or '--unregister' in sys.argv:
        import pythoncom
        if TYPELIB:
            if '--register' in sys.argv:
                tlb = os.path.abspath(
                    os.path.join(INSTALL_DIR, "typelib", "trazamed.tlb"))
                print("Registering %s" % (tlb,))
                tli = pythoncom.LoadTypeLib(tlb)
                pythoncom.RegisterTypeLib(tli, tlb)
            elif '--unregister' in sys.argv:
                k = TrazaMed
                pythoncom.UnRegisterTypeLib(k._typelib_guid_,
                                            k._typelib_version_[0],
                                            k._typelib_version_[1],
                                            0,
                                            pythoncom.SYS_WIN32)
                print("Unregistered typelib")
        import win32com.server.register
        win32com.server.register.UseCommandLine(TrazaMed)
    elif "/Automate" in sys.argv:
        # MS seems to like /automate to run the class factories.
        import win32com.server.localserver
        # win32com.server.localserver.main()
        # start the server.
        win32com.server.localserver.serve([TrazaMed._reg_clsid_])
    else:
        main()
